import numpy as np
from scipy.fft import fftn, ifftn
import sys
from typing import Self, Generator
from copy import copy
from . import brillouin_zone as brlzn
np.set_printoptions(threshold=sys.maxsize)

class WFK():
    '''
    A class for working with wavefunctions from DFT calculations

    Parameters
    ----------
    wfk_coeffs : np.ndarray
        The planewave coefficients of the wavefunction
        These should be complex values
    kpoints : np.ndarray
        A multidimensional array of 3D kpoints
        Entries along axis 0 should be individual kpoints
        Entries along axis 1 should be the kx, ky, and kz components, in that order
        The kpoints should be in reduced form
    pw_indices : np.ndarray
        Array of H, K, L indices for the planwave basis set.
        Arrays of (1,3) [H,K,L] should fill axis 0, and H, K, L values fill axis 1 in that order.
        Necessary for arranging wavefunction coefficients in 3D array.
    syrmel : np.ndarray
        A multidimensional array of 3x3 arrays of symmetry operations
    non_symm_vec : np.ndarray
        A multidimensional array of 1x3 arrays of the nonsymmorphic translation vectors for each-
        symmetry operation.
    nsym : int
        Total number of symmetry operations
    nkpt : int
        Total number of kpoints
        If a kpoints array is provided, then nkpt will be acquired the its length
    nbands : int
        Total number of bands
    ngfftx : int
        x dimension of Fourier transform grid
    ngffty : int
        y dimension of Fourier transform grid
    ngfftz : int
        z dimension of Fourier transform grid
    eigenvalues : list
        List of the eigenvalues for wavefunction at each band
        Should be ordered from least -> greatest
    fermi_energy : float
        Fermi energy 
    lattice : np.ndarray
        3x3 array containing lattice parameters
    natom : int
        Total number of atoms in unit cell
    xred : np.ndarray
        Reduced coordinates of all atoms in unit cell
        Individual atomic coordinates fill along axis 0
        X, Y, and Z components fill along axis 1, in that order
    typat : list
        Numeric labels starting from 1 and incrementing up to natom
        Order of labels should follow xred
    znucltypat : list
        List of element names
        First element of list should correspond to typat label 1, second element to label 2 and so on
    time_reversal : bool
        Select whether system has time reversal symmetry or not
        If the system is time reversal symmetric, then reciprocal space electronic states will share inversion 
        symmetry even if the real space symmetries do not include inversion
        Default assumes noncentrosymmetric systems have time reversal symmetry (True)

    Methods
    -------
    GridWFK
        Assembles plane wave coefficients on FFT grid
    RemoveGrid
        Undoes FFT grid and returns coefficients to a flat array
    FFT
        Applies Fast Fourier Transform to plane wave coefficients
    IFFT
        Applies Inverse Fast Fourier Transform to plane wave coefficients
    Normalize
        Calculates and applies normalization factor to coefficients
    Real2Reciprocal
        Calculates reciprocal lattice vectors from real space vectors
    Symmetrize
        Generates symmetrical copies from symmetry matrix operations
    SymWFK
        Generates symmetrical plane wave coefficients from operations
    XSFFormat
        Converts plane wave coefficients grid into XSF formatted grid
    RemoveXSF
        Converts XSF formatted grid into regular FFT grid
    WriteXSF
        Prints out XSF files for both the real and imaginary parts of the coefficients
    '''
    def __init__(
        self, 
        wfk_coeffs:np.ndarray=np.zeros(1), kpoints:np.ndarray=np.zeros(1), symrel:np.ndarray=np.zeros(1), 
        nsym:int=0, nkpt:int=0, nbands:int=0, ngfftx:int=0, ngffty:int=0, ngfftz:int=0, 
        eigenvalues:np.ndarray=np.zeros(1),fermi_energy:float=0.0, lattice:np.ndarray=np.zeros(1), natom:int=0, 
        xred:np.ndarray=np.zeros(1), typat:list=[], znucltypat:list=[], pw_indices:np.ndarray=np.zeros(1), 
        non_symm_vecs:np.ndarray=np.zeros(1), time_reversal:bool=True
    )->None:
        self.wfk_coeffs=wfk_coeffs
        self.kpoints=kpoints
        self.pw_indices=pw_indices
        self.symrel=symrel
        self.nsym=nsym
        self.non_symm_vecs=non_symm_vecs
        self.nkpt=nkpt
        self.nbands=nbands
        self.ngfftx=ngfftx
        self.ngffty=ngffty
        self.ngfftz=ngfftz
        self.eigenvalues=eigenvalues
        self.fermi_energy=fermi_energy
        self.lattice=lattice
        self.natom=natom
        self.xred=xred
        self.typat=typat
        self.znucltypat=znucltypat
        self.time_reversal=time_reversal
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method for putting plane wave coefficients onto 3D gridded array
    def GridWFK(
            self, band_index:int=-1
    )->Self:
        '''
        Returns copy of WFK object with coefficients in numpy 3D array grid.
        Grid is organized in (ngfftz, ngfftx, ngffty) dimensions.
        Where ngfft_ represents the _ Fourier transform grid dimension.

        Parameters
        ----------
        band_index : int
            Integer represent the band index of the wavefunction coefficients to be transformed.
            If nothing is passed, it is assumed the coefficients of a single band are supplied.
        '''
        # initialize 3D grid
        gridded_wfk = np.zeros((self.ngfftx, self.ngffty, self.ngfftz), dtype=complex)
        # update grid with wfk coefficients
        for k, kpt in enumerate(self.pw_indices):
            kx = kpt[0]
            ky = kpt[1]
            kz = kpt[2]
            if band_index >= 0:
                gridded_wfk[kx, ky, kz] = self.wfk_coeffs[band_index][k]
            else:
                gridded_wfk[kx, ky, kz] = self.wfk_coeffs[k]
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = gridded_wfk
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method for undoing grid
    def RemoveGrid(
        self, band_index:int=-1
    )->Self:
        '''
        Returns copy of WFK object with coefficients removed from the 3D gridded array.

        Parameters
        ----------
        band_index : int
            Integer represent the band index of the wavefunction coefficients to be transformed.
            If nothing is passed, it is assumed the coefficients of a single band are supplied.
        '''
        # check if coefficients are gridded before undoing grid format
        if self.wfk_coeffs.shape != (self.ngfftx,self.ngffty,self.ngfftz):
            raise ValueError((
                f'Plane wave coefficients must be in 3D grid with shape ({self.ngfftx}, {self.ngffty}, {self.ngfftz})'
                ' in order to remove the gridded format'
            ))
        if band_index >= 0:
            coeffs_no_grid = self.wfk_coeffs[band_index]
        else:
            coeffs_no_grid = self.wfk_coeffs
        # returns values at each plane wave index, undoing grid
        coeffs_no_grid = coeffs_no_grid[tuple(self.pw_indices.T)]
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = coeffs_no_grid
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method transforming reciprocal space wfks to real space
    def FFT(
            self
    )->Self:
        '''
        Returns copy of WFK with wavefunction coefficients expressed in real space.
        Assumes existing wavefunction coefficients are expressed in reciprocal space.
        '''
        # Fourier transform real grid to reciprocal grid
        reciprocal_coeffs = fftn(self.wfk_coeffs, norm='ortho')
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = np.array(reciprocal_coeffs).reshape((self.ngfftx, self.ngffty, self.ngfftz))
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method transforming real space wfks to reciprocal space
    def IFFT(
            self
    )->Self:
        '''
        Returns copy of WFK with wavefunction coefficients in expressed in reciprocal space.
        Assumes existing wavefunction coefficients are expressed in real space. 
        '''
        # Fourier transform reciprocal grid to real grid
        real_coeffs = ifftn(self.wfk_coeffs, norm='ortho')
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = np.array(real_coeffs).reshape((self.ngfftx,self.ngffty,self.ngfftz))
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method for normalizing wfks
    def Normalize(
            self
    )->Self:
        '''
        Returns copy of WFK object with normalized wavefunction coefficients such that <psi|psi> = 1.
        '''
        coeffs = np.array(self.wfk_coeffs)
        # calculate normalization constant and apply to wfk
        norm = np.dot(coeffs.flatten(), np.conj(coeffs).flatten())
        norm = np.sqrt(norm)
        new_WFK = copy(self)
        new_WFK.wfk_coeffs /= norm
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method for converting real space lattice vectors to reciprocal space vectors
    def Real2Reciprocal(
        self
    )->np.ndarray:
        '''
        Method for converting the real space lattice parameters to reciprocal lattice parameters.
        '''
        # conversion by default converts Angstrom to Bohr since ABINIT uses Bohr
        a = self.lattice[0,:]
        b = self.lattice[1,:]
        c = self.lattice[2,:]
        vol = np.dot(a,np.cross(b,c))
        b1 = 2*np.pi*(np.cross(b,c))/vol
        b2 = 2*np.pi*(np.cross(c,a))/vol
        b3 = 2*np.pi*(np.cross(a,b))/vol
        return np.array([b1,b2,b3]).reshape((3,3))
    #-----------------------------------------------------------------------------------------------------------------#
    # method for checking for time reversal symmetry
    def _CheckTimeRevSym(
        self
    ):
        if self.time_reversal:
            # if system is centrosymmetric, do not double reciprocal symmetry operations
            if -3.0 in [np.trace(mat) for mat in self.symrel]:
                self.time_reversal = False
            else:
                print((
                'Noncentrosymmetric system identified, assuming time reversal symmetry\n'
                'To change this, set time_reversal attribute to False'
                ))
    #-----------------------------------------------------------------------------------------------------------------#
    # method for finding symmetrically distinct k points
    def _FindOrbit(
        self, sym_kpts:np.ndarray
    )->tuple[list,list]:
        sym_kpts = np.round(sym_kpts, decimals=15)
        _, unique_inds = np.unique(sym_kpts, return_index=True, axis=0)
        # for each unique kpoint check original point is related by reciprocal lattice vector
        dupes = []
        for i, ind1 in enumerate(unique_inds):
            if i in dupes:
                continue
            for j, ind2 in enumerate(unique_inds):
                if i == j or j in dupes:
                    continue
                diff = np.abs(sym_kpts[ind1] - sym_kpts[ind2])
                diff[diff < 10**(-12)] = 0.0
                diff[diff > 0.999] = 1.0
                mask = np.isin(diff, np.array([0.0,1.0]))
                if mask.all():
                    dupes.append(j)
        return dupes, unique_inds.tolist()
    #-----------------------------------------------------------------------------------------------------------------#
    # function for calculating phase imparted by nonsymmorphic translation
    def _FindPhase(
        self, nonsymmvec:np.ndarray, g_vecs:np.ndarray, kpt:np.ndarray
    )->np.ndarray:
        if self.non_symm_vecs is np.zeros(1):
            return np.ones(len(g_vecs))
        elif np.sum(np.abs(nonsymmvec)) < 10**(-8):
            return np.ones(len(g_vecs))
        else:
            return np.exp(-1j*np.dot((kpt+g_vecs), nonsymmvec.T))
    #-----------------------------------------------------------------------------------------------------------------#
    # method for creating symmetrically equivalent points
    def Symmetrize(
            self, points:np.ndarray, values:np.ndarray=np.empty([]), unique:bool=True, reciprocal:bool=False,
            inverse:bool=False
    )->tuple[np.ndarray, np.ndarray]:
        '''
        Method for generating symmetric data from irreducible data.

        Parameters
        ----------
        points : np.ndarray
            Irreducible set of points.
            Shape of (N,3).
        values : np.ndarray
            Values corresponding to irreducible points (such as energy eigenvalues w/ kpoints).
            Shape of (N,1).
        unique : bool
            Check for duplicate points.
            Default is to check (True).
        reciprocal : bool
            Calculate reciprocal space symmetry matrices from real space matrices.
            Default uses real space matrices (False).
        inverse : bool
            Use inverse symmetry operations.
            Default applies forwards operation (False).
        '''
        # check if reciprocal or real space symmetries will be used
        sym_num = self.nsym
        if reciprocal:
            # nosymmorphic translations do not apply to reciprocal space
            tnons = False
            sym_mats = [np.linalg.inv(mat).T for mat in self.symrel]
            # time reversal only adds to reciprocal space symmetries
            if self.time_reversal:
                sym_mats = np.concatenate((sym_mats, [-mat for mat in sym_mats]), axis=0)
                sym_num *= 2
        else: 
            tnons = True
            sym_mats = self.symrel
        # initialize symmetrically equivalent point and value arrays
        if len(points.shape) == 1:
            points.reshape((1,points.shape[0]))
        ind_len = np.shape(points)[0]
        if values is np.empty([]):
            values = np.zeros((ind_len,1))
        sym_pts = np.zeros((sym_num*ind_len,3))
        sym_vals = np.zeros((sym_num*ind_len,self.nbands))
        if self.non_symm_vecs.all() == np.zeros(1):
            self.non_symm_vecs = np.zeros(self.nsym)
        # apply symmetry operations to points
        if inverse:
            for i, op in enumerate(sym_mats):
                if tnons:
                    points += self.non_symm_vecs[i]
                new_pts:np.ndarray = np.matmul(np.linalg.inv(op), points.T).T
                sym_pts[i*ind_len:(i+1)*ind_len,:] = new_pts
                sym_vals[i*ind_len:(i+1)*ind_len,:] = values
        else:
            for i, op in enumerate(sym_mats):
                if tnons:
                    points += self.non_symm_vecs[i]
                new_pts:np.ndarray = np.matmul(op, points.T).T
                sym_pts[i*ind_len:(i+1)*ind_len,:] = new_pts
                sym_vals[i*ind_len:(i+1)*ind_len,:] = values
        # points overlap on at edges of each symmetric block, remove duplicates
        if unique:
            dupes, unique_inds = self._FindOrbit(sym_pts)
            unique_kpts = np.array([sym_pts[ind,:] for i, ind in enumerate(unique_inds) if i not in dupes])
            unique_vals = np.array([sym_vals[ind,:] for i, ind in enumerate(unique_inds) if i not in dupes])
            return unique_kpts, unique_vals
        return sym_pts, sym_vals
    #-----------------------------------------------------------------------------------------------------------------#
    # method for creating symmetrically equivalent functions at specified kpoint
    def SymWFKs(
        self, kpoint:np.ndarray, band:int=-1
    )->Generator[Self, None, None]:
        '''
        Method for generating wavefunction planewave coefficients from coefficients of the irreducible BZ.

        Parameters
        ----------
        kpoint : np.ndarray
            A single reciprocal space point is provided to generate symmetrically equivalent coefficients.
            Shape (1,3).
        band : int
            Choose which band to pull coefficients from (indexed starting from zero).
            Default assumes coefficients from a single band are provided (-1).
        '''
        # find symmetric kpoints
        kpoint = kpoint.reshape((1,3))
        sym_kpoints, _ = self.Symmetrize(kpoint, unique=False, reciprocal=True)
        dupes, unique_inds = self._FindOrbit(sym_kpoints)
        # find symmetric planewave indices
        sym_pw_inds, _ = self.Symmetrize(self.pw_indices, unique=False, reciprocal=True)
        sym_pw_inds = sym_pw_inds.astype(int)
        ind_range = self.pw_indices.shape[0]
        # find reciprocal lattice shifts to move all points into BZ
        bz = brlzn.BZ(rec_latt=self.Real2Reciprocal())
        shifts = bz.GetShifts(sym_kpoints)
        # create WFK copies with new planewave indices
        for i, ind in enumerate(unique_inds):
            if i in dupes:
                continue
            ind1 = ind*ind_range
            ind2 = (ind+1)*ind_range
            new_pw_inds = sym_pw_inds[ind1:ind2,:]
            new_pw_inds += shifts[ind,:]
            new_coeffs = copy(self)
            new_coeffs.pw_indices = new_pw_inds
            new_coeffs.kpoints = sym_kpoints[ind,:] - shifts[ind,:]
            phase_factor = self._FindPhase(
                self.non_symm_vecs[ind % len(self.non_symm_vecs)],
                self.pw_indices,
                sym_kpoints[ind,:]
            )
            if band >= 0:
                new_coeffs.wfk_coeffs = new_coeffs.wfk_coeffs[band] * phase_factor
                yield new_coeffs
            else:
                new_coeffs.wfk_coeffs *= phase_factor
                yield new_coeffs   
    #-----------------------------------------------------------------------------------------------------------------#
    # method that returns BZ kpoints and eigenvalues
    def GetBZPtsEigs(
        self
    )->tuple[np.ndarray,np.ndarray]:
        bz = brlzn.BZ(rec_latt=self.Real2Reciprocal())
        bz_kpts, bz_eigs = self.Symmetrize(points=self.kpoints, values=self.eigenvalues, reciprocal=True)
        bz_kpts -= bz.GetShifts(bz_kpts)
        return bz_kpts, bz_eigs
    #-----------------------------------------------------------------------------------------------------------------#
    # method for expanding a grid into XSF format
    def XSFFormat(
            self
    )->Self:
        '''
        Returns copy of WFK object XSF formatted coefficients.
        Requires wfk_coeffs to be in gridded format, i.e. (ngfftz, ngfftx, ngffty) shape.
        '''
        # append zeros to ends of all axes in grid_wfk
        # zeros get replaced by values at beginning of each axis
        # this repetition is required by XSF format
        if np.shape(self.wfk_coeffs) != (self.ngfftx, self.ngffty, self.ngfftz):
            raise ValueError(
                f'''Passed array is not the correct shape:
                Expected: ({self.ngfftx}, {self.ngffty}, {self.ngfftz}),
                Received: {np.shape(self.wfk_coeffs)}
            ''')
        else:
            grid_wfk = self.wfk_coeffs
        grid_wfk = np.append(grid_wfk, np.zeros((1, self.ngffty, self.ngfftz)), axis=0)
        grid_wfk = np.append(grid_wfk, np.zeros((self.ngfftx+1, 1, self.ngfftz)), axis=1)
        grid_wfk = np.append(grid_wfk, np.zeros((self.ngfftx+1, self.ngffty+1, 1)), axis=2)
        for x in range(self.ngfftx+1):
            for y in range(self.ngffty+1):
                for z in range(self.ngfftz+1):
                    if x == self.ngfftx:
                        grid_wfk[x,y,z] = grid_wfk[0,y,z]
                    if y == self.ngffty:
                        grid_wfk[x,y,z] = grid_wfk[x,0,z]
                    if z == self.ngfftz:
                        grid_wfk[x,y,z] = grid_wfk[x,y,0]
                    if x == self.ngfftx and y == self.ngffty:
                        grid_wfk[x,y,z] = grid_wfk[0,0,z]
                    if x == self.ngfftx and z == self.ngfftz:
                        grid_wfk[x,y,z] = grid_wfk[0,y,0]
                    if z == self.ngfftz and y == self.ngffty:
                        grid_wfk[x,y,z] = grid_wfk[x,0,0]
                    if x == self.ngfftx and y == self.ngffty and z == self.ngfftz:
                        grid_wfk[x,y,z] = grid_wfk[0,0,0]
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = grid_wfk
        new_WFK.ngfftx += 1
        new_WFK.ngffty += 1
        new_WFK.ngfftz += 1
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method removing XSF formatting from density grid
    def RemoveXSF(
        self
    )->Self:
        '''
        Returns copy of WFK object without XSF formatting.
        '''
        grid = self.wfk_coeffs
        # to_be_del will be used to remove all extra data points added for XSF formatting
        to_be_del = np.ones((self.ngfftx, self.ngffty, self.ngfftz), dtype=bool)
        for z in range(self.ngfftz):
            for y in range(self.ngffty):
                for x in range(self.ngfftx):
                    # any time you reach the last density point it is a repeat of the first point
                    # remove the end points along each axis
                    if y == self.ngffty - 1 or x == self.ngfftx - 1 or z == self.ngfftz - 1:
                        to_be_del[x,y,z] = False
        # remove xsf entries from array
        grid = grid[to_be_del]
        # restore grid shape
        grid = grid.reshape((self.ngfftx-1, self.ngffty-1, self.ngfftz-1))
        new_WFK = copy(self)
        new_WFK.wfk_coeffs = grid
        new_WFK.ngfftx -= 1
        new_WFK.ngffty -= 1
        new_WFK.ngfftz -= 1
        return new_WFK
    #-----------------------------------------------------------------------------------------------------------------#
    # method for writing wavefunctions to XSF file
    def WriteXSF(
            self, xsf_file:str, _component:bool=True
    )->None:
        '''
        A method for writing numpy grids to an XSF formatted file.

        Parameters
        ----------
        xsf_file : str
            The file name.
        '''
        # first run writes out real part of eigenfunction to xsf
        if _component:
            xsf_file += '_real.xsf'
        # second run writes out imaginary part
        else:
            xsf_file += '_imag.xsf'
        with open(xsf_file, 'w') as xsf:
            print('DIM-GROUP', file=xsf)
            print('3 1', file=xsf)
            print('PRIMVEC', file=xsf)
            print(f'{self.lattice[0,0]} {self.lattice[0,1]} {self.lattice[0,2]}', file=xsf)
            print(f'{self.lattice[1,0]} {self.lattice[1,1]} {self.lattice[1,2]}', file=xsf)
            print(f'{self.lattice[2,0]} {self.lattice[2,1]} {self.lattice[2,2]}', file=xsf)
            print('PRIMCOORD', file=xsf)
            print(f'{self.natom} 1', file=xsf)
            for i, coord in enumerate(self.xred):
                atomic_num = int(self.znucltypat[self.typat[i] - 1])
                cart_coord = np.matmul(coord, self.lattice)
                print(f'{atomic_num} {cart_coord[0]} {cart_coord[1]} {cart_coord[2]}', file=xsf)
            print('ATOMS', file=xsf)
            for i, coord in enumerate(self.xred):
                atomic_num = int(self.znucltypat[self.typat[i] - 1])
                cart_coord = np.matmul(coord, self.lattice)
                print(f'{atomic_num} {cart_coord[0]} {cart_coord[1]} {cart_coord[2]}', file=xsf)
            print('BEGIN_BLOCK_DATAGRID_3D', file=xsf)
            print('datagrids', file=xsf)
            print('BEGIN_DATAGRID_3D_principal_orbital_component', file=xsf)
            print(f'{self.ngfftx} {self.ngffty} {self.ngfftz}', file=xsf)
            print('0.0 0.0 0.0', file=xsf)
            print(f'{self.lattice[0,0]} {self.lattice[0,1]} {self.lattice[0,2]}', file=xsf)
            print(f'{self.lattice[1,0]} {self.lattice[1,1]} {self.lattice[1,2]}', file=xsf)
            print(f'{self.lattice[2,0]} {self.lattice[2,1]} {self.lattice[2,2]}', file=xsf)
            count = 0
            for z in range(self.ngfftz):
                for y in range(self.ngffty):
                    for x in range(self.ngfftx):
                        count += 1
                        if _component:
                            print(self.wfk_coeffs[x,y,z].real, file=xsf, end=' ')
                        else:
                            print(self.wfk_coeffs[x,y,z].imag, file=xsf, end=' ')
                        if count == 6:
                            count = 0
                            print('\n', file=xsf, end='')
            print('END_DATAGRID_3D', file=xsf)
            print('END_BLOCK_DATAGRID_3D', file=xsf)
        # rerun method to write out imaginary part
        if _component:
            xsf_file = xsf_file.split('_real')[0]
            self.WriteXSF(xsf_file, _component=False)


import numpy as np
import pyvista as pv
from pyvistaqt import BackgroundPlotter
from copy import copy
import pickle as pkl
from matplotlib.colors import ListedColormap
from . import brillouin_zone as brlzn
from . import isosurface_class as ic
from . import wfk_class as wc
from . import translate as trslt

class Plotter():
    '''
    Class for creating 3D plot of Fermi surface and projection of BandU functions on Fermi surface.

    Parameters
    ----------
    isosurface : Isosurface
        An Isosurface object that contains the contours to be plotted
    save : bool
        Create save file of isosurface plot that can be loaded with the Load method
        Default will save the plot (True)
    save_file : str
        Name of save file
        Default is "Fermi_surface.pkl"
    empty_mesh : bool
        Allow PyVista plotter to plot meshes even if no surface is present
        Default will throw exception of an empty surface is being plotted (False)
    plot : bool
        Enable creation of PyVista Plotter
        Must be enabled for plotting, default enables plotter (True)

    Methods
    -------
    Plot
        Generates 3D plot of Fermi surface from contours of Isosurface object
    SurfaceColor
        Reads in a BandU XSF file and Fermi surface ABINIT WFK file to calculate the overlap between the BandU
        function and the states at the Fermi energy
    Load
        Loads a save file
    '''
    def __init__(
        self, isosurface:ic.Isosurface=ic.Isosurface(points=np.ones((1,3))), save:bool=True,
        empty_mesh:bool=False, _debug:bool=False, save_file:str='Fermi_surface.pkl', plot:bool=True
    ):
        self.isosurface=isosurface
        self.save=save
        self.save_file=save_file
        self._debug=_debug
        self.plot=plot
        if self.plot:
            self.p:pv.Plotter=BackgroundPlotter(window_size=(600,400))
            pv.global_theme.allow_empty_mesh=empty_mesh
            self.p.enable_depth_peeling(number_of_peels=10)
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method for turning opacity to zero outside of BZ
    def _SetOpacities(
        self, contour:pv.PolyData
    )->np.ndarray:
        bz = brlzn.BZ(self.isosurface.rec_latt)
        opacities = bz.PointLocate(contour.points, cart=False)
        opacities[opacities >= 0] = 1
        opacities[opacities < 0] = 0
        return opacities
    #-----------------------------------------------------------------------------------------------------------------#
    # method to add nesting vector to isosurface plot
    def _AddArrow(
            self, arrow:list, rec_lattice:np.ndarray, show_endpoints:bool, color:str
    )->None:
        tail = np.array(arrow[0])
        shift = np.array(arrow[1])
        tail = np.matmul(tail, rec_lattice)
        shift = np.matmul(shift, rec_lattice)
        scale = np.linalg.norm(shift).astype(float)
        py_arrow = pv.Arrow(
            start=tail,
            direction=shift,
            tip_radius=0.05/scale,
            tip_length=0.15/scale,
            shaft_radius=0.025/scale,
            scale=scale
        )
        if show_endpoints:
            points = np.array([tail, shift+tail])
            points = pv.PolyData(points)
            self.p.add_mesh(points.points, point_size=20, color='black', render_points_as_spheres=True)
        self.p.add_mesh(py_arrow, color=color)
    #-----------------------------------------------------------------------------------------------------------------#
    # method to visualize cross-section of surface
    def _CrossSection(
            self, vecs:list, points:np.ndarray, width:float, rec_lattice:np.ndarray, bz_points:np.ndarray,
            linear:bool=False, two_dim:bool=False
    )->np.ndarray:
        from scipy.spatial import Delaunay
        # cross section is defined by plane of two perpendicular vectors
        if len(vecs) == 2:
            vec1 = np.matmul(vecs[0], rec_lattice)
            vec2 = np.matmul(vecs[1], rec_lattice)
            norm = np.cross(vec1, vec2)
            norm /= np.linalg.norm(norm)
        # cross section is defined by normal vector
        else:
            vec = np.matmul(vecs, rec_lattice)
            norm = vec/np.linalg.norm(vec)
        # surface points can be thought of vectors in 3D space, here we normalize the vectors
        norm_points = np.array(points/np.linalg.norm(points, axis=1).reshape((len(points),1)))
        # find dot product between vector normal to cross section and all normalized points
        angs = np.matmul(norm, norm_points.T)
        # opacity linearly fades out as points get farther from cross section
        if linear:
            angs[angs <= width] = 2
            angs -= 1
            angs = np.abs(angs)
        # opacity is zero beyond width of cross section
        else:
            # opacity is zero beyond width on both sides of cross section
            if two_dim:
                angs = np.abs(angs)
            opacities = np.zeros(len(points))
            opacities[angs <= width] = 1
            angs = opacities  
        # set opacity to zero for points outside of the BZ
        beyond_bz = Delaunay(bz_points).find_simplex(points)
        angs[beyond_bz < 0] = 0
        return angs
    #-----------------------------------------------------------------------------------------------------------------#
    # method for plotting isosurface contours
    def Plot(
        self, show_points:bool=False, show_outline:bool=False, show_axes:bool=True, show_isosurf:bool=True,
        smooth:bool=True, lighting:bool=True, ambient:float=0.5, diffuse:float=0.5, specular:float=1.0, 
        specular_power:float=128.0, pbr:bool=False, metallic:float=0.5, roughness:float=0.5,
        colormap:str|ListedColormap='plasma', color:str='white', bz_show:bool=True, bz_width:int=3, arrow:list=[], 
        show_endpoints:bool=False, arrow_color:str='yellow', periodic_arrow:list=[], camera_position:list=[], 
        cross_section:list=[], cross_width:float=0.1, linear:bool=False, two_dim:bool=True, show_bands:float|list=1.0,
        surface_vals:list|None=None, show_ird_points:bool=False
    ):
        '''
        Method for plotting contours made from the Isosurface object

        Parameters
        ----------
        colormap : str | ListedColormap
            Choose colormap for isosurface that colors according to assigned scalars of surface\n
            Can use Colors class to use default or create custom colormaps\n
            Default is matplotlib's plasma
        bz_width : int
            Line width of Brillouin Zone\n
            Default is 3
        bz_show : bool
            Show the Brillouin Zone
            Default is to show (True)
        smooth : bool
            Use smooth lightning techniques\n
            Default is to use smoothing (True)
        lighting : bool
            Apply directional lighting to surface\n
            Default is to enable directional lighting (True)
        ambient : float
            Intensity of light on surface\n
            Default is 0.5
        diffuse : float
            Amount of light scattering\n
            Default is 0.5
        specular : float
            Amount of reflected light\n
            Default is 1.0 (max)
        specular_power : float
            Determines how sharply light is reflected\n
            Default is 128.0 (max)
        pbr : bool
            Apply physics based rendering\n
            Default is no physics based rendering (False)
        metallic : float
            Determine how metallic-looking the surface is, only considered with pbr\n
            Default is 0.5
        roughness : float
            Determine how smooth/rough surface appear, only considered with pbr\n
            Default is 0.5
        color : str
            Sets color of surface (colormap overwrites this)\n
            Sets color of reflected light (colormap does not overwrite this)\n
            Default is white
        arrow : list
            Parameters for plotting nesting vector on top of Fermi surface\n
            Element_0 of list should be starting (or tail) position of arrow\n
            Element_1 of list should be orientation of arrow with desired magnitude\n
            Both the tail and orientation should be specified in reduced reciprocal space coordinates
        arrow_color : str
            Color of nesting arrow\n
            Default is black
        show_endpoints : bool
            Plot points on the end of the arrow to make visualizing start and end easier\n
            Default is to not show endpoints (False)
        periodic_arrow : list
            Adds periodic image of arrow that is translated [X,Y,Z] cells\n
            Where X, Y, and Z are the cell indices
        show_bands : float | list
            Specifies the opacity of each band\n
            If a single float is provided, all bands will be plotted with the same opacity\n
            If a list is provided, each band will be plotted with the opacity of the respective list element
        show_axes : bool
            Plots reciprocal cell axes with a* as red, b* as green, and c* as blue\n
            Default is to show axes (True)
        cross_section : list
            Plot cross section through surface\n
            If one vector is provided, it is assumed to be the normal to the cross section plane\n
            Else, cross section is defined by plane made by two vectors\n
            Vectors should be specified in reduced coordinates
        cross_width : float
            Width of cross section\n
            Default is 0.15
        linear : bool
            Cross section linearly fades out\n
            Default is not fade out linearly (False)
        two_dim : bool
            Cross section is a 2D slice instead of a section\n
            Default is to show cross section as 2D slice (True)
        surface_vals : np.ndarray
            A list of values defining the coloration of the isosurface
            Default plots no coloration
        '''
        # save file
        if self.save:
            with open(self.save_file, 'wb') as f:
                kwargs = locals()
                kwargs.pop('self', None)
                kwargs.pop('f', None)
                pkl.dump(kwargs, f)
                pkl.dump(self.isosurface, f)
        if not self.plot:
            raise SystemExit()
        # plot BZ boundary
        vor_verts = brlzn.BZ(self.isosurface.rec_latt).vertices
        if bz_show:
            self.p.add_lines(vor_verts, color='black', width=bz_width)
        # set opacity of each band
        if type(show_bands) is float:
            band_ops = show_bands*np.ones(len(self.isosurface.contours), dtype=float)
            show_bands = band_ops.tolist()
        # loop through contours of isosurface object and plot each
        for i, contour in enumerate(self.isosurface.contours):
            if surface_vals is None:
                scalars = None
                surf_max = 0.0
            else:
                scalars = surface_vals[i]
                surf_max = np.max(surface_vals[i])
            # opacities can be adjusted to show a cross section of contours
            if cross_section != []:
                opacities = self._CrossSection(
                    cross_section,
                    contour.points, 
                    cross_width, 
                    self.isosurface.rec_latt, 
                    vor_verts,
                    linear=linear,
                    two_dim=two_dim
                )
            # otherwise opacities are set to just render contours in the BZ
            else:
                opacities = self._SetOpacities(contour=contour)
            # set opacity of bands
            opacities = [op*show_bands[i] for op in opacities] # type: ignore
            # lighting is set to make surface appear more smooth
            if smooth:
                contour = contour.smooth_taubin(n_iter=100, pass_band=0.05)
            # plot contours
            if show_isosurf:
                self.p.add_mesh(
                    contour, 
                    style='surface',
                    smooth_shading=smooth, 
                    lighting=lighting,
                    ambient=ambient,
                    diffuse=diffuse,
                    specular=specular,
                    specular_power=specular_power,
                    pbr=pbr,
                    metallic=metallic,
                    roughness=roughness,
                    scalars=scalars,
                    clim=[0.0,surf_max],
                    cmap=colormap,
                    opacity=opacities,
                    color=color,
                    show_scalar_bar=True,
                )
        # plot irreducible kpoints
        if show_ird_points:
            pts = pv.PolyData(self.isosurface.ir_kpts)
            self.p.add_mesh(pts.points, color='black')
        # plot points that are used to construct isosurface
        if show_points:
            pts = pv.PolyData(self.isosurface.points)
            self.p.add_mesh(pts.points, color='black')
        # plot outline of grid used in interpolation
        if show_outline:
            self.p.add_mesh(self.isosurface.grid.outline())
        # plot reciprocal space axes
        if show_axes:
            axes_colors = ['red', 'green', 'blue']
            for i in range(3):
                axis = self.isosurface.rec_latt[i,:]
                color = axes_colors[i]
                self._AddArrow([[0,0,0], axis], np.identity(3), False, color)
        # plot nesting vector
        if arrow != []:
            self._AddArrow(arrow, self.isosurface.rec_latt, show_endpoints, arrow_color)
        # plot a periodic image of the nesting vector
        if periodic_arrow != []:
            tail = np.array(arrow[0])
            cell = np.array(periodic_arrow, dtype=float)
            tail += cell
            arrow[0] = tail
            self._AddArrow(arrow, self.isosurface.rec_latt, show_endpoints, arrow_color)
        # set camera position
        if camera_position != []:
            camera_pos = np.array(camera_position).reshape((3,3))
            camera_pos = np.matmul(camera_pos, self.isosurface.rec_latt)
            self.p.camera_position = camera_pos
        self._Render()
    #-----------------------------------------------------------------------------------------------------------------#
    # Render isosurfaces
    def _Render(
        self
    ):
        self.p.enable_parallel_projection() # type: ignore
        self.p.enable_custom_trackball_style(
            left='rotate',
            shift_left='spin',
            right='pan',
        ) # type: ignore
        self.p.set_focus([0.0,0.0,0.0]) # type: ignore
        self.p.show()
        self.p.app.exec_() # type: ignore
    #-----------------------------------------------------------------------------------------------------------------#
    # method to compute bandu fxn and one electron wfk overlaps
    def _OverlapsWithSym(
        self, ir_wfk:wc.WFK, bandu:wc.WFK
    )->np.ndarray:
        # kpoint to symmetrically generate
        kpt = ir_wfk.kpoints
        # find number of distinct symmetries 
        sym_kpoints, _ = ir_wfk.Symmetrize(kpt, unique=False, reciprocal=True)
        dupes, unique_inds = ir_wfk._FindOrbit(sym_kpoints)
        count = sum([1 for i, _ in enumerate(unique_inds) if i not in dupes])
        # initialize array for overlap values
        num_bands = len(self.isosurface.nbands)
        overlap_vals = np.zeros((count,num_bands), dtype=float)
        # loop over bands
        for i, band in enumerate(self.isosurface.nbands):
            # generate symmetric coefficients for each band
            for j, wfk in enumerate(ir_wfk.SymWFKs(kpoint=kpt, band=band)):
                wfk = wfk.GridWFK()
                wfk = wfk.IFFT()
                wfk = wfk.Normalize()
                overlap = np.sum(np.conj(bandu.wfk_coeffs)*wfk.wfk_coeffs)
                overlap = np.square(np.abs(overlap))
                overlap_vals[j,i] = overlap
        return overlap_vals
    #-----------------------------------------------------------------------------------------------------------------#
    # method to compute bandu fxn and one electron wfk overlaps
    def _OverlapsNoSym(
        self, ir_wfk:wc.WFK, bandu:wc.WFK
    )->np.ndarray:
        # initialize array for overlap values
        num_bands = len(self.isosurface.nbands)
        overlap_vals = np.zeros((1,num_bands), dtype=float)
        # loop over bands
        for i, band in enumerate(self.isosurface.nbands):
            wfk = ir_wfk.GridWFK(band_index=band)
            wfk = wfk.IFFT()
            wfk = wfk.Normalize()
            overlap = np.sum(np.conj(bandu.wfk_coeffs)*wfk.wfk_coeffs)
            overlap = np.square(np.abs(overlap))
            overlap_vals[0,i] = overlap
        return overlap_vals
    #-----------------------------------------------------------------------------------------------------------------#
    # method to interpolate overlap values
    def _InterpolateOverlaps(
        self, contour:pv.PolyData, overlap_values:np.ndarray
    )->np.ndarray:
        # recreate eigenvalue grid as base for surface color grid
        color_grid = copy(self.isosurface.grid)
        # translate overlap values to 3x3x3 grid
        trans_color_points, trans_color_values = trslt.TranslatePoints(
            self.isosurface.points,
            overlap_values.reshape((-1,1)),
            self.isosurface.rec_latt
        )
        # construct PyVista object from translated overlap grid
        trans_color_points = pv.PolyData(trans_color_points)
        trans_color_points['values'] = trans_color_values
        # interpolate translated grid
        color_grid = color_grid.interpolate(
            trans_color_points,
            sharpness=2.0,
            radius=self.isosurface.radius,
            strategy='null_value'
        )
        # sample color values from interpolated color grid
        color_sample = contour.sample(color_grid)
        return color_sample.active_scalars
    #-----------------------------------------------------------------------------------------------------------------#
    # method to calculate surface color values from XSF and WFK
    def SurfaceColor(
        self, wfk_path:str, xsf_path:str, sym:bool=False
    )->list:
        '''
        Method for calculating BandU function overlap with states at a specified isoenergy.
        Requires BandU XSF file and ABINIT WFK file.

        Parameters
        ----------
        wfk_path : str
            Path to ABINIT WFK file
        xsf_path : str
            Path to BandU XSF file
        sym : bool
            Symmetrically generate full Brillouin Zone
            Default will not generate Brillouin Zone and instead symmetrize surface colors (False)
        '''
        from . import abinit_reader as ar
        from . import xsf_reader as xsfr
        # list of overlap values
        overlaps = np.zeros(1)
        # read fermi surface wavefunction
        fermi_wfk = ar.AbinitWFK(filename=wfk_path)   
        # get number of bands
        nbands = len(self.isosurface.nbands)
        # paths to real and imaginary bandu xsf files
        real_path = xsf_path + '_real.xsf'
        imag_path = xsf_path + '_imag.xsf'
        # read in xsf
        real_fxn = xsfr.XSF(xsf_file=real_path)
        imag_fxn = xsfr.XSF(xsf_file=imag_path)
        print('XSF read')
        # convert xsf to wfk object
        bandu_fxn = wc.WFK(
            wfk_coeffs=real_fxn.ReadGrid() + 1j*imag_fxn.ReadGrid(),
            ngfftx=real_fxn.ngfftx,
            ngffty=real_fxn.ngffty,
            ngfftz=real_fxn.ngfftz
        )
        # remove xsf format
        bandu_fxn = bandu_fxn.RemoveXSF()
        # loop through fermi surface kpoints and calc overlap with bandu fxn
        for i, kpt in enumerate(fermi_wfk.ReadWFK()):
            if sym:
                vals = self._OverlapsWithSym(kpt, bandu_fxn)
            else:
                vals = self._OverlapsNoSym(kpt, bandu_fxn)
            if i == 0:
                overlaps = vals
            else:
                overlaps = np.concatenate((overlaps, vals), axis=0)
        # symmetrically permute overlap values if they were only calculated on irreducible BZ wedge when sym==False
        if not sym:
            symrel = fermi_wfk.symrel
            nsym = fermi_wfk.nsym
            kpts = fermi_wfk.kpts
            all_kpts = np.zeros((1,3))
            all_overlaps = np.zeros((1,nbands))
            new_wfk = wc.WFK(symrel=np.array(symrel), nsym=nsym, nbands=nbands)
            for i, kpt in enumerate(kpts):
                unique_kpts, _ = new_wfk.Symmetrize(
                    points=kpt,
                    reciprocal=True
                )
                all_kpts = np.concatenate((all_kpts, unique_kpts), axis=0)
                new_overlaps = np.repeat(overlaps[i,:].reshape((1,-1)), unique_kpts.shape[0], axis=0)
                all_overlaps = np.concatenate((all_overlaps, new_overlaps), axis=0)
            kpts = np.delete(all_kpts, 0, axis=0)
            overlaps = np.delete(all_overlaps, 0, axis=0)
            self.isosurface.points = np.matmul(kpts, self.isosurface.rec_latt)
        # interpolate overlap values for smooth coloration
        scalars = []
        for i in range(overlaps.shape[1]):
            interp_vals = self._InterpolateOverlaps(self.isosurface.contours[i], overlaps[:,i])
            scalars.append(interp_vals)
        return scalars
    #-----------------------------------------------------------------------------------------------------------------#
    # method to load save file
    def Load(
        self, save_path:str|list, **kwargs
    ):
        '''
        Method for loading saved contours 

        Parameters
        ----------
        colormap : str | ListedColormap
            Choose colormap for isosurface that colors according to assigned scalars of surface\n
            Can use Colors class to use default or create custom colormaps\n
            Default is matplotlib's plasma
        bz_width : int
            Line width of Brillouin Zone\n
            Default is 3
        bz_show : bool
            Show the Brillouin Zone
            Default is to show (True)
        smooth : bool
            Use smooth lightning techniques\n
            Default is to use smoothing (True)
        lighting : bool
            Apply directional lighting to surface\n
            Default is to enable directional lighting (True)
        ambient : float
            Intensity of light on surface\n
            Default is 0.5
        diffuse : float
            Amount of light scattering\n
            Default is 0.5
        specular : float
            Amount of reflected light\n
            Default is 1.0 (max)
        specular_power : float
            Determines how sharply light is reflected\n
            Default is 128.0 (max)
        pbr : bool
            Apply physics based rendering\n
            Default is no physics based rendering (False)
        metallic : float
            Determine how metallic-looking the surface is, only considered with pbr\n
            Default is 0.5
        roughness : float
            Determine how smooth/rough surface appear, only considered with pbr\n
            Default is 0.5
        color : str
            Sets color of surface (colormap overwrites this)\n
            Sets color of reflected light (colormap does not overwrite this)\n
            Default is white
        arrow : list
            Parameters for plotting nesting vector on top of Fermi surface\n
            Element_0 of list should be starting (or tail) position of arrow\n
            Element_1 of list should be orientation of arrow with desired magnitude\n
            Both the tail and orientation should be specified in reduced reciprocal space coordinates
        arrow_color : str
            Color of nesting arrow\n
            Default is black
        show_endpoints : bool
            Plot points on the end of the arrow to make visualizing start and end easier\n
            Default is to not show endpoints (False)
        periodic_arrow : list
            Adds periodic image of arrow that is translated [X,Y,Z] cells\n
            Where X, Y, and Z are the cell indices
        show_bands : float | list
            Specifies the opacity of each band\n
            If a single float is provided, all bands will be plotted with the same opacity\n
            If a list is provided, each band will be plotted with the opacity of the respective list element
        show_axes : bool
            Plots reciprocal cell axes with a* as red, b* as green, and c* as blue\n
            Default is to show axes (True)
        cross_section : list
            Plot cross section through surface\n
            If one vector is provided, it is assumed to be the normal to the cross section plane\n
            Else, cross section is defined by plane made by two vectors\n
            Vectors should be specified in reduced coordinates
        cross_width : float
            Width of cross section\n
            Default is 0.15
        linear : bool
            Cross section linearly fades out\n
            Default is not fade out linearly (False)
        two_dim : bool
            Cross section is a 2D slice instead of a section\n
            Default is to show cross section as 2D slice (True)
        surface_vals : np.ndarray
            A list of values defining the coloration of the isosurface
            Default plots no coloration
        '''
        # functionality for adding multiple projections onto one surface
        if type(save_path) == list:
            all_scalars = []
            for save in save_path:
                with open(save, 'rb') as f:
                    kwargs_dict:dict = pkl.load(f)
                    self.isosurface:ic.Isosurface = pkl.load(f)
                all_scalars.append(kwargs_dict['surface_vals'])
            sizes = []
            for surface_vals in all_scalars[0]:
                sizes.append(surface_vals.shape)
            new_scalars = [np.zeros((size)) for size in sizes]
            for scalars in all_scalars:
                for i, surface_vals in enumerate(scalars):
                    new_scalars[i] += surface_vals
            kwargs_dict['surface_vals'] = new_scalars # type: ignore
        # replot a single surface
        else:
            with open(save_path, 'rb') as f: # type: ignore
                kwargs_dict:dict = pkl.load(f)
                self.isosurface:ic.Isosurface = pkl.load(f)
        for k, val in kwargs.items(): # type: ignore
            kwargs_dict[k] = val # type: ignore       
        self.save = False
        self.plot = True
        self.Plot(**kwargs_dict) # type: ignore


import pyvista as pv
import numpy as np
from . import translate as trnslt
from . import abinit_reader as ar
from . import wfk_class as wc
from . import brillouin_zone as brlzn

# object for creating reciprocal space energy isosurfaces
class Isosurface():
    '''
    Class for constructing contours from WFK eigenvalues

    Parameters
    ----------
    points : np.ndarray
        Array of reciprocal space points to construct contour over
        Shape (N,3)
    values : np.ndarray
        Array of eigenvalues for each reciprocal space point
        Shape (N,1) or (N,#Bands)
    nbands : list[int]
        List of bands to create contours for, aka indices of bands that cross energy isosurface
    rec_latt : np.ndarray
        Reciprocal lattice vectors
    wfk_name : str
        If points and values arrays are not provided, then construct contours from reading in an ABINIT WFK file
        Path to wfk file
    grid_steps : tuple
        Set fineness of grid that contours are interpolated on
        Default (50,50,50)
    fermi_energy : float
        Fermi energy
        Default 0.0
    energy_level : float
        Sets energy isosurface relative to the Fermi energy
        Default 0.0 (this samples the Fermi energy)
    width : float
        Sets the range above and below the energy_level that eigenvalues are sampled from
        Default 0.005 (Assumes units of Hartree) 
    radius : float
        Radius of gaussian interpolation
        Default 0.05, this scales inversely with kpoint grid size (larger grids require smaller radius)
    sym_ops : np.ndarray
        Array of symmetry operations for generating points in Brillouin Zone
    grid_stretch : float
        Value that stretches grid that is used to interpolate bands
        If portions of Brillouin Zone are cropped off, increase this value
        Default is 0.15

    Methods
    -------
    Contour
        Generates energy isosurface contours
    '''
    def __init__(
        self, points:np.ndarray=np.zeros(1), values:np.ndarray=np.zeros(1), rec_latt:np.ndarray=np.zeros(1), 
        wfk_name:str='', grid_steps:tuple=(50,50,50), fermi_energy:float=0.0, energy_level:float=0.0, 
        width:float=0.005, radius:float=0.05, nbands:list[int]=[], sym_ops:np.ndarray=np.zeros(1), 
        grid_stretch:float=0.15
    )->None:
        # define attributes
        self.points=points
        self.values=values
        self.wfk_name=wfk_name
        self.rec_latt=rec_latt
        self.fermi_energy=fermi_energy
        self.energy_level=energy_level
        self.width=width
        self.radius=radius
        self.nbands=nbands
        self.sym_ops=sym_ops
        self.grid_stretch=grid_stretch
        # check if enough information is provided to class to construct isosurfaces
        if self.points.shape == (1,) and self.values.shape == (1,):
            if wfk_name == '':
                raise ValueError((
                    'Either the points and the values attributes must be defined or the wfk_name attribute '
                    'must be defined'
                ))
            else:
                self.points, self.values, self.nbands, self.ir_kpts = self._ReadAbinit(wfk_name)
        # check if information provided is in correct format
        if len(self.values.shape) > 2:
            raise ValueError((
                f'Provided values have dimension {len(self.values.shape)}'
                'Values must be 1 or 2 dimensional'
            ))
        else:
            self._valdim=len(self.values.shape)
        # more attributes, these rely on points and values attributes
        self.contours:list[pv.PolyData] = []
        self.grid_steps=grid_steps
        self.grid = self._MakeGrid()
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method for creating grid to interplote on
    def _MakeGrid(
            self
    )->pv.ImageData:
        '''
        Method for constructing PyVista grid used in interpolation prior to plotting
        *This method assumes your points are in Cartesian format*

        Parameters
        ----------
        points : np.ndarry
            Numpy array with shape (N, 3) where N is number of points\n
            These are the points that lie within your energy range, generally from GetValAndKpt method\n
            These are NOT the points of the grid
        steps : tuple
            Define the number of grid points along (x,y,z)\n
            Default is 50 points along each axis\n
            Spacing of grid is automatically calculated from steps and points
        '''
        # helper functions for finding grid origin and setting grid spacing
        def _GetMax(pts:np.ndarray):
            xmax = pts[:,0].max()   
            ymax = pts[:,1].max()
            zmax = pts[:,2].max()
            return xmax, ymax, zmax
        def _GetMin(pts:np.ndarray):
            xmin = pts[:,0].min()   
            ymin = pts[:,1].min()
            zmin = pts[:,2].min()
            return xmin, ymin, zmin
        # begin method for making grid
        xmax, ymax, zmax = _GetMax(self.points)
        xmin, ymin, zmin = _GetMin(self.points)
        dimx = self.grid_steps[0]
        dimy = self.grid_steps[1]
        dimz = self.grid_steps[2]
        grid = pv.ImageData()
        stretch_factor = self.grid_stretch
        dim_stretch = [2*stretch_factor/dim for dim in self.grid_steps]
        grid.origin = (
            xmin - 0.5*stretch_factor, 
            ymin - 0.5*stretch_factor,
            zmin - 0.5*stretch_factor
        )
        grid.spacing = (
            2*xmax/dimx + dim_stretch[0], 
            2*ymax/dimy + dim_stretch[1], 
            2*zmax/dimz + dim_stretch[2]
        )
        grid.dimensions = (dimx, dimy, dimz)
        return grid
    #-----------------------------------------------------------------------------------------------------------------#
    # method for selecting which contour to perform depending on if a single band or many bands are passed
    def Contour(
        self
    ):
        if self._valdim == 1:
            self._Contour(points=self.points, values=self.values)
        else:
            for band in self.values.T:
                self._Contour(points=self.points, values=band)
    #-----------------------------------------------------------------------------------------------------------------#
    # method for drawing contour from list of values
    def _Contour(
        self, points:np.ndarray, values:np.ndarray
    ):
        # setup grid interpolation
        null_value = self.fermi_energy + self.energy_level + self.width/2 * 1.05
        trans_pts, trans_vals = trnslt.TranslatePoints(
            points=points, 
            values=values.reshape((values.shape[0],1)), 
            lattice_vecs=self.rec_latt
        )
        trans_pts = pv.PolyData(trans_pts)
        trans_pts['values'] = trans_vals
        inter_grid:pv.ImageData = self.grid.interpolate(
            trans_pts, 
            radius=self.radius, 
            sharpness=2.0, 
            strategy='null_value', 
            null_value=null_value
        )
        # create contour
        iso_range = [self.fermi_energy + self.energy_level, self.fermi_energy + self.energy_level]
        contour:pv.PolyData = inter_grid.contour(
            isosurfaces=2,
            rng=iso_range,
            method='contour'
        )
        self.contours.append(contour)
    #-----------------------------------------------------------------------------------------------------------------#
    # method for reading kpoints and eigenvalues from ABINIT WFK file
    def _ReadAbinit(
        self, filename:str
    )->tuple[np.ndarray,np.ndarray, list, np.ndarray]:
        # collect kpoints and eigenvalues from wfk file
        wfk = ar.AbinitWFK(filename=filename)
        eigs = []
        syms = np.array(wfk.symrel)
        nsym = wfk.nsym
        kpoints = np.array(wfk.kpts)
        ir_kpts = kpoints
        nbands = wfk.bands[0]
        nkpt = wfk.nkpt
        self.fermi_energy = wfk.fermi
        self.rec_latt = wc.WFK(lattice=wfk.real_lattice).Real2Reciprocal()
        for wfk_obj in wfk.ReadEigenvalues():
            eigs.append(wfk_obj.eigenvalues)
        eigs = np.array(eigs).reshape((nkpt,nbands))
        # look through eigenvalues to find which bands to contour
        min_val = self.fermi_energy + self.energy_level - self.width/2
        max_val = self.fermi_energy + self.energy_level + self.width/2
        band_num = []
        for i in range(nbands):
            for eigval in eigs[:,i]:
                if min_val <= eigval <= max_val:
                    band_num.append(i)
                    break            
        # return kpoints and bands of interest
        bands = np.take(eigs, band_num, axis=1)
        shifts = brlzn.BZ(self.rec_latt).GetShifts(kpoints)
        kpoints = kpoints - shifts
        ir_kpts -= shifts
        ir_kpts = np.matmul(ir_kpts,self.rec_latt)
        wfk = wc.WFK(symrel=syms, nsym=nsym, nbands=len(band_num), nkpt=nkpt)
        all_kpts = np.zeros((1,3))
        all_eigs = np.zeros((1,bands.shape[1]))
        # symmetrize kpoints
        for i, kpt in enumerate(kpoints):
            unique_kpts, _ = wfk.Symmetrize(
                points=kpt,
                reciprocal=True,
            )
            all_kpts = np.concatenate((all_kpts, unique_kpts), axis=0)
            new_eigs = np.repeat(bands[i,:].reshape((1,-1)), unique_kpts.shape[0], axis=0)
            all_eigs = np.concatenate((all_eigs, new_eigs), axis=0)
        kpoints = np.delete(all_kpts, 0, axis=0)
        bands = np.delete(all_eigs, 0, axis=0)
        kpoints = np.matmul(kpoints, self.rec_latt)
        return kpoints, bands, band_num, ir_kpts


import numpy as np
from scipy.spatial import Voronoi, KDTree, Delaunay
from . import translate as trnslt

# Brillouin Zone object
class BZ():
    def __init__(
        self, rec_latt:np.ndarray
    )->None:
        self.rec_latt = rec_latt
        _3xgrid, _ = trnslt.TranslatePoints(np.zeros((1,3)), np.zeros(1), self.rec_latt)
        self._3xgrid = _3xgrid
        self.vertices = self._BZpts()
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # Method used to check if nearest points all occur in same plane
    def _CheckPts(
        self, pts:np.ndarray, verts:np.ndarray
    )->bool:
        # get cartesian coordinates
        verts = np.take(verts, pts, axis=0)
        # rounding avoids floating point errors in if comparison below
        verts = np.round(verts, decimals=8)
        # last point is the furthest, check if it is in same plane as other points
        pt = verts[-1,:]
        check = False
        # check if points are in same plane
        if pt[0] == verts[0,:][0] and pt[0] == verts[1,:][0] and pt[0] == verts[2,:][0]:
            check = True 
        if pt[1] == verts[0,:][1] and pt[1] == verts[1,:][1] and pt[1] == verts[2,:][1]:
            check = True
        if pt[2] == verts[0,:][2] and pt[2] == verts[1,:][2] and pt[2] == verts[2,:][2]:
            check = True
        return check
    #-----------------------------------------------------------------------------------------------------------------#
    def _BZpts(
        self
    )->np.ndarray:
        # Voronoi tessallate the lattice points
        vor_tessellation = Voronoi(self._3xgrid)
        # find Voronoi region that encapsulates Brillouin zone
        vor_cell = -1
        for i, region in enumerate(vor_tessellation.regions):
            if -1 not in region and region != []:
                vor_cell = i
                break
        if vor_cell < 0:
            raise ValueError(
                'Unable to construct Brillouin Zone from reciprocal lattice, check lattice parameters.'
            )
        # get vertices of BZ Voronoi region
        vor_cell_verts = vor_tessellation.regions[vor_cell]
        vor_verts = np.take(vor_tessellation.vertices, vor_cell_verts, axis=0)
        # find which vertices are closest to each other
        point_cloud = KDTree(vor_verts)
        vert_inds = []
        for i, vert in enumerate(vor_verts):
            _, nearest_pts = point_cloud.query(vert, k=4)
            c = 4
            # iterate check until points are no longer in same plane
            while self._CheckPts(nearest_pts, vor_verts):
                c += 1
                _, nearest_pts = point_cloud.query(vert, k=c)
                len_arr = c - 4
                arr = -1*np.arange(len_arr) - 2
                nearest_pts = np.delete(nearest_pts, arr, axis=0)
            # current index is a repeated entry every other point since this is how PyVista interprets lines
            nearest_pts = np.insert(nearest_pts, [2,4], [i,i])
            vert_inds.append(nearest_pts)
        # convert nested list to flat list
        vert_inds = [int(ind) for verts in vert_inds for ind in verts]
        # get cartesian coordinates
        vor_verts = np.take(vor_verts, vert_inds, axis=0)
        return vor_verts
    #-----------------------------------------------------------------------------------------------------------------#
    # method for determining shifts required to translate points outside BZ back into BZ
    def GetShifts(
        self, points:np.ndarray, cart:bool=True
    )->np.ndarray:
        '''
        Method for calculating necessary reciprocal lattice shift to move point into Brillouin Zone.

        Parameters
        ----------
        points : np.ndarray
            List of points to calculate shifts for.
        cart : bool
            Convert points from reduced to Cartesian coordinates.
            Default will apply conversion (True).
        '''
        # make points mulitdimensional if not already
        if len(points.shape) == 1:
            points = points.reshape((1,points.shape[0]))
        # convert to cartesian
        if cart:
            points = self.MakeCart(points=points)
        # find points outside of BZ
        outside_pts = self.PointLocate(points)
        # calculate shifts to move outside points back into BZ
        shifts = np.zeros((outside_pts[outside_pts < 0].shape[0],3))
        rec_grid = KDTree(self._3xgrid)
        shift_grid, _ = trnslt.TranslatePoints(np.zeros((1,3)), np.zeros(1), np.identity(3))
        for i, pt in enumerate(points[outside_pts < 0]):
            _, closest_pt = rec_grid.query(pt, k=1)
            shifts[i,:] = shift_grid[closest_pt, :]
        all_shifts = np.zeros((points.shape[0],3))
        all_shifts[outside_pts < 0] = shifts
        return all_shifts.astype(int)
    #-----------------------------------------------------------------------------------------------------------------#
    # method for finding if points are outside or inside of BZ
    def PointLocate(
        self, points:np.ndarray, cart:bool=True
    )->np.ndarray:
        '''
        Method for finding which points are outside, on the edge, and inside the Brillouin Zone.

        Parameters
        ----------
        points : np.ndarray
            Array of points to be checked against the Brillouin Zone, must be in cartesian format
            Shape (N,3)

        Returns
        -------
        np.ndarray
        elements < 0 correspond to points outside the Brillouin Zone
        elements > 0 correspond to points inside
        '''
        if cart:
            points = self.MakeCart(points=points)
        outside_pts = Delaunay(self.vertices).find_simplex(points)
        return outside_pts
    #-----------------------------------------------------------------------------------------------------------------#
    # method for finding if point is on BZ edge/face
    # not working
    def _BZEdgePt(
        self, points:np.ndarray, cart:bool=True
    )->np.ndarray:
        shift_grid, _ = trnslt.TranslatePoints(np.zeros((1,3)), np.zeros(1), np.identity(3))
        test_point = self.PointLocate(shift_grid + points, cart=cart)
        return test_point
    #-----------------------------------------------------------------------------------------------------------------#
    # method for converting points to cartesian format
    def MakeCart(
        self, points:np.ndarray
    )->np.ndarray:
        '''
        Method for converting points from reduced to cartesian format

        Parameters
        ----------
        points : np.ndarray
            Array of points to be converted
            Shape (N,3)

        Returns
        -------
        np.ndarray
        points in cartesian format
        Shape (N,3)
        '''
        cart_pts = np.matmul(points, self.rec_latt)
        return cart_pts
    #-----------------------------------------------------------------------------------------------------------------#
    # method for converting points to reduced format
    def MakeRed(
        self, points:np.ndarray
    )->np.ndarray:
        '''
        Method for converting points from cartesian to reduced format

        Parameters
        ----------
        points : np.ndarray
            Array of points to be converted
            Shape (N,3)

        Returns
        -------
        np.ndarray
        points in reduced format
        Shape (N,3)
        '''
        red_pts = np.matmul(points, np.linalg.inv(self.rec_latt))
        return red_pts


import numpy as np
from typing import Generator
from copy import copy
from . import brillouin_zone as brlzn
from . import translate as trnslt
from . import wfk_class as wc

class BandU():
    def __init__(
        self, wfks:Generator, energy_level:float, width:float, grid:bool=True, fft:bool=True, norm:bool=True,
        sym:bool=True, low_mem:bool=False, plot:bool=True
    )->None:
        '''
        BandU object with methods for finding states and computing BandU functions from states.

        Parameters
        ----------
        wfks : Generator
            An iterable generator of WFK objects with wavefunction coefficients, k-points, and eigenvalue attributes.
        energy_level : float
            The energy level of interest relative to the Fermi energy.
        width : float
            Defines how far above and below the energy_level is searched for states.
            Search is done width/2 above and below, so total states captured are within 'width' energy.
        grid : bool
            Determines whether or not wavefunction coefficients are converted to 3D numpy grid.
            Default converts to grid (True).
        fft : bool
            Determines whether or not wavefunction coefficients are Fourier transformed to real space.
            Default converts from reciprocal space to real space (True).
        norm : bool
            Determines whether or not wavefunction coefficients are normalized.
            Default normalizes coefficients (True)
        plot : bool
            Plots eigenvalues from principal component analysis.
            Default will save plot (True)
        low_mem : bool
            Run the program on a lower memory setting
            The low_mem tag will print plane wave cofficients to a Python pickle to read from disk later.
            Default does not run in low memory mode (False)

        Methods
        -------
        ToXSF
            Writes real and imaginary parts of BandU functions to XSF files
        '''
        self.grid:bool=grid
        self.fft:bool=fft
        self.norm:bool=norm
        self.sym:bool=sym
        self.low_mem:bool=low_mem
        self.found_states:int=0
        self.bandu_fxns:list[wc.WFK]=[]
        self.plot=plot
        # find all states within width
        self._FindStates(energy_level, width, wfks)
        print(f'{self.found_states} states found within specified energy range')
        # construct principal orbital components
        principal_vals = self._PrincipalComponents()
        # plot eigenvalues from PCA
        if plot:
            self._PlotEigs(principal_vals)
        # normalize bandu functions
        for i in range(self.found_states):
            self.bandu_fxns[i] = self.bandu_fxns[i].Normalize()
        # compute ratios
        omega_vals, omega_check = self._CheckOmega()
        # write output file
        fermi = self.bandu_fxns[0].fermi_energy
        with open('eigenvalues.out', 'w') as f:
            print(f'Width: {width}, found states: {self.found_states}', file=f)
            print(f'Energy level: {energy_level+fermi}, Fermi energy: {fermi}', file=f)
            print(np.abs(principal_vals), file=f)   
            print('Omega Values', file=f)
            print(omega_vals, file=f)   
            print('Omega Check (value of 0 indicates Omega is at local extremum)', file=f)
            print(omega_check, file=f)
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method transforming reciprocal space wfks to real space
    def _FindStates(
        self, energy_level:float, width:float, wfks:Generator[wc.WFK,None,None]
    ):
        # loop through every state
        for state in wfks:
            # check if state has a band that crosses the width
            min_en = state.fermi_energy + energy_level - width/2
            max_en = state.fermi_energy + energy_level + width/2
            for i, band in enumerate(state.eigenvalues):
                if min_en <= band <= max_en:
                    self.found_states += 1
                    coeffs = copy(state)
                    coeffs.wfk_coeffs = coeffs.wfk_coeffs[i]
                    for wfk in self._Process(coeffs):
                        self.bandu_fxns.append(wfk)
        if self.bandu_fxns is []:
            raise ValueError(
            '''Identified 0 states within provided width.
            Action: Increase width or increase fineness of kpoint grid.
            ''')
    #-----------------------------------------------------------------------------------------------------------------#
    # method for processing planewave coefficient data from FindStates
    def _Process(
            self, state:wc.WFK
    )->Generator[wc.WFK,None,None]:
        funcs:list[wc.WFK] = []
        # generate symmetrically equivalent coefficients
        if self.sym:
            for sym_coeffs in state.SymWFKs(kpoint=state.kpoints):
                self.found_states += 1
                funcs.append(sym_coeffs)
            self.found_states -= 1
        else:
            # shift point back into Brillouin Zone as necessary
            rec_latt = state.Real2Reciprocal()
            shift = brlzn.BZ(rec_latt=rec_latt).GetShifts(state.kpoints)
            state.pw_indices += shift
            state.kpoints = state.kpoints.reshape((-1,3)) - shift
            funcs.append(state)
        # apply desired transformations
        for wfk in funcs:
            if self.grid:
                wfk = wfk.GridWFK()
            if self.fft:
                wfk = wfk.IFFT()
            if self.norm:
                wfk = wfk.Normalize()
            yield wfk
    #-----------------------------------------------------------------------------------------------------------------#
    # check if states along BZ edge have been collected
    # --not functional--
    def _CheckEdgeCase(
        self
    ):
        bz = brlzn.BZ(self.bandu_fxns[0].Real2Reciprocal())
        x = self.bandu_fxns[0].ngfftx
        y = self.bandu_fxns[0].ngffty
        z = self.bandu_fxns[0].ngfftz
        all_shifts:list[np.ndarray] = []
        wfk_to_shift:list[int] = []
        dupes:list[wc.WFK] = []
        for i, wfk in enumerate(self.bandu_fxns):
            # if point on edge, translate to find other periodic points
            edge_pts = bz._BZEdgePt(wfk.kpoints)
            print(wfk.kpoints)
            print(edge_pts)
            if len(edge_pts[edge_pts > 0]) > 0:
                shifts, _ = trnslt.TranslatePoints(np.zeros((1,3)), np.zeros(1), np.identity(3))
                shifts = shifts[edge_pts > 0].astype(int)
                all_shifts.append(shifts)
                wfk_to_shift.append(i)
        for i in wfk_to_shift:
            for shifts in all_shifts:
                for shift in shifts:
                    shifted_wfk = copy(self.bandu_fxns[i])
                    shifted_wfk = shifted_wfk.RemoveGrid()
                    shifted_wfk.pw_indices += shift
                    shifted_wfk = shifted_wfk.GridWFK()
                    shifted_wfk.wfk_coeffs = shifted_wfk.wfk_coeffs.reshape((1,x*y*z))
                    shifted_dupe = copy(self.bandu_fxns)
                    shifted_dupe[i] = shifted_wfk
                    dupes.extend(shifted_dupe)
                    self.duped_states += self.found_states
        self.bandu_fxns.extend(dupes)
        #-----------------------------------------------------------------------------------------------------------------#
    # find principal components
    def _PrincipalComponents(
        self
    )->np.ndarray:
        total_states = self.found_states
        # organize wfk coefficients 
        x = self.bandu_fxns[0].ngfftx
        y = self.bandu_fxns[0].ngffty
        z = self.bandu_fxns[0].ngfftz
        mat = np.zeros((total_states,x*y*z), dtype=complex)
        for i in range(total_states):
            mat[i,:] = self.bandu_fxns[i].wfk_coeffs.reshape((1,x*y*z))
        # compute overlap matrix
        print('Computing overlap matrix')
        overlap_mat = np.matmul(np.conj(mat), mat.T)
        # diagonlize matrix
        principal_vals, principal_vecs = np.linalg.eig(overlap_mat)
        principal_vecs = principal_vecs.T
        # organize eigenvectors and eigenvalues
        sorted_inds = np.flip(principal_vals.argsort())
        principal_vals = np.take(principal_vals, sorted_inds)
        principal_vecs = np.take(principal_vecs, sorted_inds, axis=0)
        mat = np.matmul(principal_vecs, mat)
        for i in range(total_states):
            self.bandu_fxns[i].wfk_coeffs = mat[i,:]
        return principal_vals 
    #-----------------------------------------------------------------------------------------------------------------#
    # find ratio of real and imaginary components
    def _CheckOmega(
        self
    )->tuple[np.ndarray, np.ndarray]:
        total_states = self.found_states
        omega_vals = np.zeros((total_states, 3), dtype=float)
        vals = np.linspace(start=-0.01, stop=0.01, num=3)
        for i, val in enumerate(vals):
            for j in range(total_states):
                coeffs:np.ndarray = copy(self.bandu_fxns[j].wfk_coeffs)
                coeffs *= np.exp(1j*val*np.pi)
                omega = np.sum(coeffs.real*coeffs)/np.sum(coeffs.imag*coeffs)
                omega = np.abs(omega)
                omega_vals[j,i] = omega
        omega_diff1 = (omega_vals[:,1] - omega_vals[:,0])
        omega_diff2 = (omega_vals[:,2] - omega_vals[:,1])
        omega_check = np.sign(omega_diff1) + np.sign(omega_diff2)
        return omega_vals[:,1], omega_check
    #-----------------------------------------------------------------------------------------------------------------#
    # plot eigenvalues from PCA
    def _PlotEigs(
        self, eigvals:np.ndarray
    ):
        import matplotlib.pyplot as plt
        x = np.arange(self.found_states) + 1
        y = np.abs(eigvals.flatten())
        figsize = (12,6)
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot()
        ax.plot(
            x, 
            y, 
            color='black', 
            linestyle='-', 
            marker='o', 
            mfc='red', 
            markersize=8
        )
        ax.spines[['right','top']].set_visible(False)
        ax.tick_params(axis='both', labelsize=12)
        plt.xlim(1.0,len(y)+5.0)
        plt.ylim=(0,np.max(y))
        mod_val = round(self.found_states/5 - 0.5)
        plt.xticks(ticks=[val for val in range(0,len(y)+1) if val % mod_val == 0])
        plt.rcParams['font.family'] = 'Times New Roman'
        plt.savefig('bandu_eigenvalues.png',dpi=500)
    #-----------------------------------------------------------------------------------------------------------------#
    # make xsf of BandU functions
    def ToXSF(
        self, nums:list[int]=[], xsf_name:str='Principal_orbital_component'
    ):
        total_states = self.found_states
        if nums is []:
            nums = [1,total_states]
        else:
            # check if list has only 2 elements
            if len(nums) != 2:
                raise ValueError(f'nums should contain two values, {len(nums)} were received.')
            # check if function number is within defined range
            if nums[0] < 1:
                print('First element of nums cannot be lower than 1, changing to 1 now.')
                nums[0] = 1
            # update function number list if it exceeds maximum number of bandu functions
            if nums[1] > total_states:
                print(f'Printing up to max Band-U function number: {total_states}')
                nums[1] = total_states
            # check if lower limit is within defined range
            if nums[0] > nums[1]:
                nums[0] = nums[1]
        print(f'Writing XSF files for Band-U functions {nums[0]} through {nums[1]}.')
        # write xsf files
        x = self.bandu_fxns[0].ngfftx
        y = self.bandu_fxns[0].ngffty
        z = self.bandu_fxns[0].ngfftz
        for i in range(nums[0]-1, nums[1]):
            file_name = xsf_name + f'_{i+1}'
            wfk = copy(self.bandu_fxns[i])
            wfk.wfk_coeffs = wfk.wfk_coeffs.reshape((x,y,z))
            wfk = wfk.XSFFormat()
            wfk.WriteXSF(xsf_file=file_name)


import numpy as np

atom_labels = {1:'H', 2:'He', 3:'Li', 4:'Be', 5:'B', 6:'C', 7:'N', 8:'O', 9:'F', 10:'Ne', 11:'Na', 12:'Mg', 13:'Al', 
               14:'Si', 15:'P', 16:'S', 17:'Cl', 18:'Ar', 19:'K', 20:'Ca', 21:'Sc', 22:'Ti', 23:'V', 24:'Cr', 25:'Mn',
               26:'Fe', 27:'Co', 28:'Ni', 29:'Cu', 30:'Zn', 31:'Ga', 32:'Ge', 33:'As', 34:'Se', 35:'Br', 36:'Kr', 
               37:'Rb', 38:'Sr', 39:'Y', 40:'Zr', 41:'Nb', 42:'Mo', 43:'Tc', 44:'Ru', 45:'Rh', 46:'Pd', 47:'Ag', 
               48:'Cd', 49:'In', 50:'Sn', 51:'Sb', 52:'Te', 53:'I', 54:'Xe', 55:'Cs', 56:'Ba', 57:'La', 58:'Ce', 
               59:'Pr', 60:'Nd', 61:'Pm', 62:'Sm', 63:'Eu', 64:'Gd', 65:'Tb', 66:'Dy', 67:'Ho', 68:'Er', 69:'Tm', 
               70:'Yb', 71:'Lu', 72:'Hf', 73:'Ta', 74:'W', 75:'Re', 76:'Os', 77:'Ir', 78:'Pt', 79:'Au', 80:'Hg', 
               81:'Tl', 82:'Pb', 83:'Bi', 84:'Po', 85:'At', 86:'Rn', 87:'Fr', 88:'Ra', 89:'Ac', 90:'Th', 91:'Pa', 
               92:'U'}

class XSF():
    def __init__(
            self, 
            xsf_file:str='WFK.xsf'
        )->None:
    # if xsf file is supplied, read in parameters
        self.xsf_file = xsf_file
        with open(xsf_file, 'r') as xsf:
            self.xsf_lines = xsf.readlines()
        self.lattice = np.zeros((3,3))
        for i, line in enumerate(self.xsf_lines):
            # get lattice vectors from XSF file
            if line.strip() == 'PRIMVEC':
                self.lattice[0,:] = [float(val) for val in self.xsf_lines[i+1].strip().split(' ')]
                self.lattice[1,:] = [float(val) for val in self.xsf_lines[i+2].strip().split(' ')]
                self.lattice[2,:] = [float(val) for val in self.xsf_lines[i+3].strip().split(' ')]
            # get number of atoms and atomic symbols
            # get atomic coordinates from XSF file
            if line.strip() == 'PRIMCOORD':
                self.natoms = int(self.xsf_lines[i+1].strip().split(' ')[0])
                self.coords = np.zeros((self.natoms, 3))
                self.elements = []
                for atom in range(self.natoms):
                    coord = self.xsf_lines[i+atom+2].strip().split(' ')
                    coord = [float(val) for val in coord]
                    element = atom_labels[int(coord[0])]
                    self.elements.append(element)
                    del coord[0]
                    self.coords[atom,:] = coord
            # once density block is reached, get ngfft spacing and end init
            if line.strip() == 'BEGIN_DATAGRID_3D_principal_orbital_component':
                ngfft_spacing = self.xsf_lines[i+1].strip().split(' ')
                ngfft_spacing = [int(val) for val in ngfft_spacing]
                self.ngfftx = ngfft_spacing[0]
                self.ngffty = ngfft_spacing[1]
                self.ngfftz = ngfft_spacing[2]
                return None
    #-----------------------------------------------------------------------------------------------------------------#
    # method reading in BandU eigenfunction from XSF
    def ReadGrid(
        self
    )->np.ndarray:
        '''
        Method for reading in density grid from XSF file. Returns grid as N dimensional numpy array.

        Parameters
        ----------
        undo_xsf : bool
            If True (which is the default), then the XSF formatting is undone by removing end points.
        '''
        density_lines:list|np.ndarray=[]
        for i, line in enumerate(self.xsf_lines):
            # get density block, this assumes density is the end most data grid in the XSF
            if line.strip() == 'BEGIN_DATAGRID_3D_principal_orbital_component':
                # density starts 6 lines down from BEGIN_DATAGRID_3D_principal_orbital_component header
                density_lines = self.xsf_lines[i+6:]
                # last line indicates end of data block, remove it
                del density_lines[-1]
                # last entry in datagrid has a string indicating end of grid, remove it
                last_line = density_lines[-1].strip().split(' ')
                last_line = [val for val in last_line if val != 'END_DATAGRID_3D']
                # cast line back to a single string
                last_line = ' '.join(last_line)
                density_lines[-1] = last_line
        if density_lines is []:
            raise LookupError('3D grid data not found in XSF file')
        # convert density to 3D array of floats
        density_lines = [line.strip().split(' ') for line in density_lines]
        density_lines = [val for line in density_lines for val in line if val != '']
        density_lines = np.array(density_lines, dtype=float)
        density_lines = density_lines.reshape((self.ngfftx, self.ngffty, self.ngfftz), order='F')
        return density_lines


import numpy as np

def TranslatePoints(
        points:np.ndarray, values:np.ndarray, lattice_vecs:np.ndarray
)->tuple[np.ndarray, np.ndarray]:
    npts = len(points)
    # translate points along each lattice direction to ensure grid space is filled
    trans_points = np.zeros((27*npts,3))
    trans_values = np.zeros((27*npts,1))
    for i in range(3):
        even_ind = 9*i
        odd_ind = 9*i+1 
        # translate points along z axis
        layer_pts = points + (i-1)*lattice_vecs[2,:]
        trans_points[even_ind*npts:(even_ind+1)*npts,:] = layer_pts
        # translate along +x and -x
        trans_points[odd_ind*npts:(odd_ind+1)*npts,:] = layer_pts + lattice_vecs[0,:]
        trans_points[(even_ind+2)*npts:(even_ind+3)*npts,:] = layer_pts - lattice_vecs[0,:]
        # along +y and -y
        trans_points[(odd_ind+2)*npts:(odd_ind+3)*npts,:] = layer_pts + lattice_vecs[1,:]
        trans_points[(even_ind+4)*npts:(even_ind+5)*npts,:] = layer_pts - lattice_vecs[1,:]
        # along +x+y and +x-y
        trans_points[(odd_ind+4)*npts:(odd_ind+5)*npts,:] = layer_pts + lattice_vecs[0,:] + lattice_vecs[1,:]
        trans_points[(even_ind+6)*npts:(even_ind+7)*npts,:] = layer_pts - lattice_vecs[0,:] + lattice_vecs[1,:]
        # along -x+y and -x-y
        trans_points[(odd_ind+6)*npts:(odd_ind+7)*npts,:] = layer_pts + lattice_vecs[0,:] - lattice_vecs[1,:]
        trans_points[(even_ind+8)*npts:(even_ind+9)*npts,:] = layer_pts - lattice_vecs[0,:] - lattice_vecs[1,:]
        # repeat translations for values array
        trans_values[even_ind*npts:(even_ind+1)*npts,:] = values
        trans_values[odd_ind*npts:(odd_ind+1)*npts,:] = values
        trans_values[(even_ind+2)*npts:(even_ind+3)*npts,:] = values
        trans_values[(odd_ind+2)*npts:(odd_ind+3)*npts,:] = values
        trans_values[(even_ind+4)*npts:(even_ind+5)*npts,:] = values
        trans_values[(odd_ind+4)*npts:(odd_ind+5)*npts,:] = values
        trans_values[(even_ind+6)*npts:(even_ind+7)*npts,:] = values
        trans_values[(odd_ind+6)*npts:(odd_ind+7)*npts,:] = values
        trans_values[(even_ind+8)*npts:(even_ind+9)*npts,:] = values
    return trans_points, trans_values


import numpy as np
from matplotlib.colors import ListedColormap

# class for creating custom colormaps for isosurface
class Colors():
    #-----------------------------------------------------------------------------------------------------------------#
    # define a few colormaps for easy plotting
    def __init__(self):
        self.reds = self.Colormap([np.array([125, 125, 125])/255, 
                                   np.array([255, 46, 31])/255
        ])
        self.stoplight = self.Colormap([np.array([255, 46, 31])/255, 
                                        np.array([255, 180, 31])/255, 
                                        np.array([26, 209, 23])/255
        ])
        self.WhBlRd = self.Colormap([np.array([240, 240, 240])/255, 
                                     np.array([0, 0, 255])/255, 
                                     np.array([255, 0, 0])/255
        ])
        self.PlTlPk = self.Colormap([np.array([74, 2, 115])/255, 
                                     np.array([11, 217, 210])/255, 
                                     np.array([255, 0, 255])/255
        ])
        self.blues = self.Colormap([np.array([240, 240, 240])/255, 
                                     np.array([75, 222, 222])/255, 
                                     np.array([5, 47, 255])/255
        ])
    #-----------------------------------------------------------------------------------------------------------------#
    # method to create custom colormap from list of rgb codes
    def Colormap(self, colorlist:list)->ListedColormap:
        num = 250*(len(colorlist)-1)
        parts = 250
        colors = np.empty((num, 4))
        reds = np.zeros(num)
        greens = np.zeros(num)
        blues = np.zeros(num)
        for i, color in enumerate(colorlist):
            if i == len(colorlist) - 1:
                break
            reds[i*parts:(i+1)*parts] += np.linspace(color[0], colorlist[i+1][0], parts)
            greens[i*parts:(i+1)*parts] += np.linspace(color[1], colorlist[i+1][1], parts)
            blues[i*parts:(i+1)*parts] += np.linspace(color[2], colorlist[i+1][2], parts)
        colors[:,0] = reds
        colors[:,1] = greens
        colors[:,2] = blues
        colors[:,3] = np.ones(num)
        return ListedColormap(colors)


import struct
import numpy as np
from typing import Generator, Union
import sys
from . import wfk_class as wc
np.set_printoptions(threshold=sys.maxsize)

def bytes2float(bin_data)->float:
    return struct.unpack('<d', bin_data)[0]

def bytes2int(bin_data)->int:
    return int.from_bytes(bin_data, 'little', signed=True)

class Abinit7WFK():
    def __init__(
        self, filename:str
    ):
        self.filename = filename
        self._ReadHeader()
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method to read wavefunction header for abinit version 7
    def _ReadHeader(
        self
    )->None:
        wfk = open(self.filename, 'rb')
        print('Reading WFK header')
        #---------------#
        # unpack integers
        self.header = bytes2int(wfk.read(4))
        self.version = wfk.read(6).decode()
        self.headform = bytes2int(wfk.read(4))
        self.fform = bytes2int(wfk.read(4))
        wfk.read(8)
        self.bandtot = bytes2int(wfk.read(4))
        self.date = bytes2int(wfk.read(4))
        self.intxc = bytes2int(wfk.read(4))
        self.ixc = bytes2int(wfk.read(4))
        self.natom = bytes2int(wfk.read(4))
        self.ngfftx = bytes2int(wfk.read(4))
        self.ngffty = bytes2int(wfk.read(4))
        self.ngfftz = bytes2int(wfk.read(4))
        self.nkpt = bytes2int(wfk.read(4))
        self.nspden = bytes2int(wfk.read(4))
        self.nspinor = bytes2int(wfk.read(4))
        self.nsppol = bytes2int(wfk.read(4))
        self.nsym = bytes2int(wfk.read(4))
        self.npsp = bytes2int(wfk.read(4))
        self.ntypat = bytes2int(wfk.read(4))
        self.occopt = bytes2int(wfk.read(4))
        self.pertcase = bytes2int(wfk.read(4))
        self.usepaw = bytes2int(wfk.read(4))
        if self.usepaw == 1:
            raise NotImplementedError(
                f'''PAW potentials are not supported by this program for Abinit v{self.version}.
                This program supports PAW potentials in Abinit v10, consider upgrading to the latest version.
                '''
            )
        #--------------#
        # unpack doubles
        self.ecut = bytes2float(wfk.read(8))
        self.ecutdg = bytes2float(wfk.read(8))
        self.ecutsm = bytes2float(wfk.read(8))
        self.ecut_eff = bytes2float(wfk.read(8))
        self.qptnx = bytes2float(wfk.read(8))
        self.qptny = bytes2float(wfk.read(8))
        self.qptnz = bytes2float(wfk.read(8))
        rprimd_ax = bytes2float(wfk.read(8))
        rprimd_ay = bytes2float(wfk.read(8))
        rprimd_az = bytes2float(wfk.read(8))
        rprimd_bx = bytes2float(wfk.read(8))
        rprimd_by = bytes2float(wfk.read(8))
        rprimd_bz = bytes2float(wfk.read(8))
        rprimd_cx = bytes2float(wfk.read(8))
        rprimd_cy = bytes2float(wfk.read(8))
        rprimd_cz = bytes2float(wfk.read(8))
        #------------------------#
        # convert Bohr to Angstrom
        self.real_lattice = 0.529177*np.array([[rprimd_ax, rprimd_ay, rprimd_az],
                                    [rprimd_bx, rprimd_by, rprimd_bz], 
                                    [rprimd_cx, rprimd_cy, rprimd_cz]]).reshape((3,3))
        self.stmbias = bytes2float(wfk.read(8))
        self.tphysel = bytes2float(wfk.read(8))
        self.tsmear = bytes2float(wfk.read(8))
        #---------------#
        # unpack integers
        self.usewvl = bytes2int(wfk.read(4))
        wfk.read(8)
        self.istwfk:list[int] = []
        for i in range(self.nkpt):
            val = bytes2int(wfk.read(4))
            self.istwfk.append(val)    
        self.bands:list[int] = []
        for i in range(self.nkpt*self.nsppol):
            val = bytes2int(wfk.read(4))
            self.bands.append(val)
        self.npwarr:list[int] = []
        for i in range(self.nkpt):
            val = bytes2int(wfk.read(4))
            self.npwarr.append(val)
        self.so_psp:list[int] = []
        for i in range(self.npsp):
            val = bytes2int(wfk.read(4))
            self.so_psp.append(val)
        self.symafm:list[int] = []
        for i in range(self.nsym):
            val = bytes2int(wfk.read(4))
            self.symafm.append(val)
        self.symrel:list[np.ndarray] = []
        for i in range(self.nsym):
            arr = np.zeros((3,3))
            arr[0,0] = bytes2int(wfk.read(4))
            arr[1,0] = bytes2int(wfk.read(4))
            arr[2,0] = bytes2int(wfk.read(4))
            arr[0,1] = bytes2int(wfk.read(4))
            arr[1,1] = bytes2int(wfk.read(4))
            arr[2,1] = bytes2int(wfk.read(4))
            arr[0,2] = bytes2int(wfk.read(4))
            arr[1,2] = bytes2int(wfk.read(4))
            arr[2,2] = bytes2int(wfk.read(4))
            self.symrel.append(arr)
        self.typat:list[int] = []
        for i in range(self.natom):
            val = bytes2int(wfk.read(4))
            self.typat.append(val)
        #--------------#
        # unpack doubles
        self.kpts:list[np.ndarray] = []
        for i in range(self.nkpt):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.kpts.append(vec)
        self.occ:list[float] = []
        for i in range(self.bandtot):
            val = bytes2float(wfk.read(8))
            self.occ.append(val)
        self.tnons:list[np.ndarray] = []
        for i in range(self.nsym):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.tnons.append(vec)
        self.znucltypat:list[float] = []
        for i in range(self.ntypat):
            val = bytes2float(wfk.read(8))
            self.znucltypat.append(val)
        self.wtk:list[float] = []           
        for i in range(self.nkpt):
            val = bytes2float(wfk.read(8))
            self.wtk.append(val)
        #-----------------------#
        # unpack pseudopotentials
        wfk.read(4)
        self.title:list[str] = []
        self.znuclpsp:list[float] = []
        self.zionpsp:list[float] = []
        self.pspso:list[int] = []
        self.pspdat:list[int] = []
        self.pspcod:list[int] = []
        self.pspxc:list[int] = []
        self.lmn_size:list[int] = []
        for i in range(self.npsp):
            wfk.read(4)
            self.title.append(wfk.read(132).decode())
            self.znuclpsp.append(bytes2float(wfk.read(8)))
            self.zionpsp.append(bytes2float(wfk.read(8)))
            self.pspso.append(bytes2int(wfk.read(4)))
            self.pspdat.append(bytes2int(wfk.read(4)))
            self.pspcod.append(bytes2int(wfk.read(4)))
            self.pspxc.append(bytes2int(wfk.read(4)))
            self.lmn_size.append(bytes2int(wfk.read(4)))
            wfk.read(4)
        #------------------#
        # unpack coordinates
        wfk.read(4)
        self.residm = bytes2float(wfk.read(8))
        self.xred:list[np.ndarray] = []
        for i in range(self.natom):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.xred.append(vec)
        #------------------------------------#
        # unpack total energy and fermi energy
        self.etotal = bytes2float(wfk.read(8))
        self.fermi = bytes2float(wfk.read(8))
        wfk.read(4)
        print('WFK header read')
        wfk.close()
    #-----------------------------------------------------------------------------------------------------------------#
    # check for time reversal symmetry
    def _CheckTimeRev(
        self
    ):
        # if system is centrosymmetric, do not double reciprocal symmetry operations
        if True in [np.array_equal(-np.identity(3),mat) for mat in self.symrel]:
            self.time_reversal = False
        else:
            self.time_reversal = True
            print((
            'Noncentrosymmetric system identified, assuming time reversal symmetry\n'
            'To change this, set "check_time_rev" keyword to False when calling ReadWFK() or ReadEigenvalues()'
            ))
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read entire body of abinit version 7 wavefunction file
    def ReadWFK(
        self, energy_level=np.nan, width=np.nan, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        '''
        Method that constructs WFK objects from ABINIT v7 WFK file

        Parameters
        ----------
        energy_level : float
        The energy level to pull plane wave coefficients from
        This is relative to the Fermi energy
        If not defined but width is defined, default is 0 Hartree

        width : float
        Defines range about energy_level to search for plane wave coefficients
        Total range is equal to width, so look width/2 above and below energy_level
        If not defined but energy_level is defined, default is 0.005 Hartree
        '''
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        # read wfk file
        wfk = open(self.filename, 'rb')
        skip = True
        if energy_level is np.nan:
            if width is np.nan:
                skip = False
            else:
                energy_level = 0
        else:
            if width is np.nan:
                width = 0.005
        #-----------#
        # skip header
        wfk.read(298)
        wfk.read(4*(2*self.nkpt + self.nkpt*self.nsppol + self.npsp + 10*self.nsym + self.natom))
        wfk.read(8*(4*self.nkpt + self.bandtot + 3*self.nsym + self.ntypat + 3*self.natom))
        wfk.read(self.npsp*(176))
        #-------------------------------#
        # begin reading wavefunction body
        for i in range(self.nsppol):
            for j in range(self.nkpt):
                print(f'Reading kpoint {j+1} of {self.nkpt}', end='\r')
                if j+1 == self.nkpt:
                    print('\n', end='')
                pw_indices:list[tuple] = []
                eigenvalues:list[float] = []
                wfk.read(4)
                npw = bytes2int(wfk.read(4))
                self.nspinor = bytes2int(wfk.read(4))
                nband_temp = bytes2int(wfk.read(4))
                wfk.read(8)
                for pw in range(npw):
                    kx = bytes2int(wfk.read(4))
                    ky = bytes2int(wfk.read(4))
                    kz = bytes2int(wfk.read(4))
                    pw_indices.append((kx, ky, kz))
                wfk.read(8)
                for nband in range(nband_temp):
                    eigenval = bytes2float(wfk.read(8))
                    eigenvalues.append(eigenval)
                # skip reading coefficients if they energy level of interest is not in band
                if skip:
                    min_val = self.fermi + energy_level - width/2
                    max_val = self.fermi + energy_level + width/2
                    eig_found = False
                    for eigval in eigenvalues:
                        if min_val <= eigval <= max_val:
                            eig_found = True
                            break
                    if eig_found:
                        yield self._ReadCoeffs(
                            wfk,
                            nband_temp,
                            npw,
                            eigenvalues,
                            pw_indices,
                            ind=j
                        )
                    else:
                        wfk.read(nband_temp*8)
                        wfk.read(4)
                        wfk.read(nband_temp*(8 + npw*16))
                else:
                    yield self._ReadCoeffs(
                        wfk,
                        nband_temp,
                        npw,
                        eigenvalues,
                        pw_indices,
                        ind=j
                    )
        print('WFK body read')
        wfk.close()
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read in plane wave coefficients
    def _ReadCoeffs(
        self, wfk_file, nbands, npws, eigs, pw_inds, ind
    )->wc.WFK:
        occupancies = np.zeros((1,nbands), dtype=float)
        coeffs = np.zeros((nbands,npws), dtype=complex)
        for nband in range(nbands):
            occ = bytes2float(wfk_file.read(8))
            occupancies[:,nband] = occ
        wfk_file.read(4)
        for nband in range(nbands):
            wfk_file.read(4)
            cg:np.ndarray = np.zeros((1,npws), dtype=complex)
            for pw in range(npws):
                cg1 = bytes2float(wfk_file.read(8))
                cg2 = bytes2float(wfk_file.read(8))
                cg[:,pw] = cg1 + 1j*cg2
            coeffs[nband,:] = cg
            wfk_file.read(4) 
        return wc.WFK(
            eigenvalues=np.array(eigs), 
            wfk_coeffs=np.array(coeffs),
            pw_indices=np.array(pw_inds),
            kpoints=np.array(self.kpts[ind]),
            nkpt=self.nkpt,
            nbands=self.bands[0],
            ngfftx=self.ngfftx,
            ngffty=self.ngffty,
            ngfftz=self.ngfftz,
            symrel=np.array(self.symrel),
            nsym=self.nsym,
            lattice=self.real_lattice,
            natom=self.natom,
            xred=np.array(self.xred),
            typat=self.typat,
            znucltypat=self.znucltypat,
            fermi_energy=self.fermi,
            non_symm_vecs=np.array(self.tnons),
            time_reversal=self.time_reversal
        )
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read only eigenvalues from body of abinit version 7 wavefunction file
    def ReadEigenvalues(
        self, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        '''
        Method that constructs WFK objects from ABINIT v7 WFK file.
        '''
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        # read wfk
        wfk = open(self.filename, 'rb')
        #-----------#
        # skip header
        wfk.read(298)
        wfk.read(4*(2*self.nkpt + self.nkpt*self.nsppol + self.npsp + 10*self.nsym + self.natom))
        wfk.read(8*(4*self.nkpt + self.bandtot + 3*self.nsym + self.ntypat + 3*self.natom))
        wfk.read(self.npsp*(176))
        #-------------------------------#
        # begin reading wavefunction body
        for i in range(self.nsppol):
            for j in range(self.nkpt):
                print(f'Reading kpoint {j+1} of {self.nkpt}', end='\r')
                if j+1 == self.nkpt:
                    print('\n', end='')
                eigenvalues:list[float] = []
                wfk.read(4)
                npw = bytes2int(wfk.read(4))
                self.nspinor = bytes2int(wfk.read(4))
                nband_temp = bytes2int(wfk.read(4))
                wfk.read(8)
                wfk.read(12*npw)
                wfk.read(8)
                #------------------------------------------------------------------#
                # only need eigenvalues for Fermi surface, skip over everything else
                for nband in range(nband_temp):
                    eigenval = bytes2float(wfk.read(8))
                    eigenvalues.append(eigenval)   
                wfk.read(nband_temp*8)
                wfk.read(4)
                wfk.read(nband_temp*(8 + npw*16))
                yield wc.WFK(
                    eigenvalues=np.array(eigenvalues), 
                    kpoints=np.array(self.kpts[j]),
                    nkpt=self.nkpt,
                    nbands=self.bands[0],
                    ngfftx=self.ngfftx,
                    ngffty=self.ngffty,
                    ngfftz=self.ngfftz,
                    symrel=np.array(self.symrel),
                    nsym=self.nsym,
                    lattice=self.real_lattice,
                    natom=self.natom,
                    xred=np.array(self.xred),
                    typat=self.typat,
                    znucltypat=self.znucltypat,
                    fermi_energy=self.fermi,
                    non_symm_vecs=np.array(self.tnons),
                    time_reversal=self.time_reversal
                )
        print('WFK body read')
        wfk.close()

#---------------------------------------------------------------------------------------------------------------------#
#----------------------------------------------------- END CLASS -----------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#

class Abinit10WFK():
    def __init__(
        self, filename:str
    ):
        self.filename = filename
        self._ReadHeader()
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # method to read wavefunction header for abinit version 7
    def _ReadHeader(
        self
    )->None:
        wfk = open(self.filename, 'rb')
        print('Reading WFK header')
        #---------------#
        # unpack integers
        self.header = bytes2int(wfk.read(4))
        self.version = wfk.read(6).decode()
        wfk.read(2)
        self.headform = bytes2int(wfk.read(4))
        self.fform = bytes2int(wfk.read(4))
        wfk.read(8)
        self.bandtot = bytes2int(wfk.read(4))
        self.date = bytes2int(wfk.read(4))
        self.intxc = bytes2int(wfk.read(4))
        self.ixc = bytes2int(wfk.read(4))
        self.natom = bytes2int(wfk.read(4))
        self.ngfftx = bytes2int(wfk.read(4))
        self.ngffty = bytes2int(wfk.read(4))
        self.ngfftz = bytes2int(wfk.read(4))
        self.nkpt = bytes2int(wfk.read(4))
        self.nspden = bytes2int(wfk.read(4))
        self.nspinor = bytes2int(wfk.read(4))
        self.nsppol = bytes2int(wfk.read(4))
        self.nsym = bytes2int(wfk.read(4))
        self.npsp = bytes2int(wfk.read(4))
        self.ntypat = bytes2int(wfk.read(4))
        self.occopt = bytes2int(wfk.read(4))
        self.pertcase = bytes2int(wfk.read(4))
        self.usepaw = bytes2int(wfk.read(4))
        #--------------#
        # unpack doubles
        self.ecut = bytes2float(wfk.read(8))
        self.ecutdg = bytes2float(wfk.read(8))
        self.ecutsm = bytes2float(wfk.read(8))
        self.ecut_eff = bytes2float(wfk.read(8))
        self.qptnx = bytes2float(wfk.read(8))
        self.qptny = bytes2float(wfk.read(8))
        self.qptnz = bytes2float(wfk.read(8))
        rprimd_ax = bytes2float(wfk.read(8))
        rprimd_ay = bytes2float(wfk.read(8))
        rprimd_az = bytes2float(wfk.read(8))
        rprimd_bx = bytes2float(wfk.read(8))
        rprimd_by = bytes2float(wfk.read(8))
        rprimd_bz = bytes2float(wfk.read(8))
        rprimd_cx = bytes2float(wfk.read(8))
        rprimd_cy = bytes2float(wfk.read(8))
        rprimd_cz = bytes2float(wfk.read(8))
        #------------------------#
        # convert Bohr to Angstrom
        self.real_lattice = 0.529177*np.array([
            [rprimd_ax, rprimd_ay, rprimd_az],
            [rprimd_bx, rprimd_by, rprimd_bz], 
            [rprimd_cx, rprimd_cy, rprimd_cz]
        ]).reshape((3,3))
        self.stmbias = bytes2float(wfk.read(8))
        self.tphysel = bytes2float(wfk.read(8))
        self.tsmear = bytes2float(wfk.read(8))
        #---------------#
        # unpack integers
        self.usewvl = bytes2int(wfk.read(4))
        self.nshiftk_orig = bytes2int(wfk.read(4))
        self.nshiftk = bytes2int(wfk.read(4))
        self.mband = bytes2int(wfk.read(4))
        wfk.read(8)
        self.istwfk:list[int] = []
        for _ in range(self.nkpt):
            val = bytes2int(wfk.read(4))
            self.istwfk.append(val)      
        self.bands:list[int] = []
        for _ in range(self.nkpt*self.nsppol):
            val = bytes2int(wfk.read(4))
            self.bands.append(val)
        self.npwarr:list[int] = []
        for _ in range(self.nkpt):
            val = bytes2int(wfk.read(4))
            self.npwarr.append(val)
        self.so_psp:list[int] = []
        for _ in range(self.npsp):
            val = bytes2int(wfk.read(4))
            self.so_psp.append(val)
        self.symafm:list[int] = []
        for _ in range(self.nsym):
            val = bytes2int(wfk.read(4))
            self.symafm.append(val)
        self.symrel:list[np.ndarray] = []
        for _ in range(self.nsym):
            arr = np.zeros((3,3))
            arr[0,0] = bytes2int(wfk.read(4))
            arr[1,0] = bytes2int(wfk.read(4))
            arr[2,0] = bytes2int(wfk.read(4))
            arr[0,1] = bytes2int(wfk.read(4))
            arr[1,1] = bytes2int(wfk.read(4))
            arr[2,1] = bytes2int(wfk.read(4))
            arr[0,2] = bytes2int(wfk.read(4))
            arr[1,2] = bytes2int(wfk.read(4))
            arr[2,2] = bytes2int(wfk.read(4))
            self.symrel.append(arr)
        self.typat:list[int] = []
        for _ in range(self.natom):
            val = bytes2int(wfk.read(4))
            self.typat.append(val)
        #--------------#
        # unpack doubles
        self.kpts:list[np.ndarray] = []
        for _ in range(self.nkpt):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.kpts.append(vec)
        self.occ:list[float] = []
        for _ in range(self.bandtot):
            val = bytes2float(wfk.read(8))
            self.occ.append(val)
        self.tnons:list[np.ndarray] = []
        for _ in range(self.nsym):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.tnons.append(vec)
        self.znucltypat:list[float] = []
        for _ in range(self.ntypat):
            val = bytes2float(wfk.read(8))
            self.znucltypat.append(val)
        self.wtk:list[float] = []           
        for _ in range(self.nkpt):
            val = bytes2float(wfk.read(8))
            self.wtk.append(val)
        #------------------#
        # unpack coordinates
        wfk.read(8)
        self.residm = bytes2float(wfk.read(8))
        self.xred:list[np.ndarray] = []
        for _ in range(self.natom):
            vec = np.zeros(3)
            vec[0] = bytes2float(wfk.read(8))
            vec[1] = bytes2float(wfk.read(8))
            vec[2] = bytes2float(wfk.read(8))
            self.xred.append(vec)
        #------------------------------------#
        # unpack total energy and fermi energy
        self.etotal = bytes2float(wfk.read(8))
        self.fermi = bytes2float(wfk.read(8))
        #-------------#
        # unpack floats
        self.amu:list[float] = []
        for _ in range(self.ntypat):
            val = bytes2float(wfk.read(8))
            self.amu.append(val)
        #----------------------------#
        # unpack reciprocal space info
        wfk.read(8)
        self.kptopt = bytes2int(wfk.read(4))
        self.pawcpxocc = bytes2int(wfk.read(4))
        self.nelect = bytes2float(wfk.read(8))
        self.cellcharge = bytes2float(wfk.read(8))
        self.icoulomb = bytes2int(wfk.read(4))
        rec_latt_ax = bytes2int(wfk.read(4))
        rec_latt_ay = bytes2int(wfk.read(4))
        rec_latt_az = bytes2int(wfk.read(4))
        rec_latt_bx = bytes2int(wfk.read(4))
        rec_latt_by = bytes2int(wfk.read(4))
        rec_latt_bz = bytes2int(wfk.read(4))
        rec_latt_cx = bytes2int(wfk.read(4))
        rec_latt_cy = bytes2int(wfk.read(4))
        rec_latt_cz = bytes2int(wfk.read(4))
        self.kptr_latt = np.array([
            [rec_latt_ax, rec_latt_ay, rec_latt_az],
            [rec_latt_bx, rec_latt_by, rec_latt_bz],
            [rec_latt_cx, rec_latt_cy, rec_latt_cz]
        ]).reshape((3,3))
        rec_latt_ax = bytes2int(wfk.read(4))
        rec_latt_ay = bytes2int(wfk.read(4))
        rec_latt_az = bytes2int(wfk.read(4))
        rec_latt_bx = bytes2int(wfk.read(4))
        rec_latt_by = bytes2int(wfk.read(4))
        rec_latt_bz = bytes2int(wfk.read(4))
        rec_latt_cx = bytes2int(wfk.read(4))
        rec_latt_cy = bytes2int(wfk.read(4))
        rec_latt_cz = bytes2int(wfk.read(4))
        self.kptr_latt_orig = np.array([
            [rec_latt_ax, rec_latt_ay, rec_latt_az],
            [rec_latt_bx, rec_latt_by, rec_latt_bz],
            [rec_latt_cx, rec_latt_cy, rec_latt_cz]
        ]).reshape((3,3))
        shift_orig_x = bytes2float(wfk.read(8))
        shift_orig_y = bytes2float(wfk.read(8))
        shift_orig_z = bytes2float(wfk.read(8))
        self.origin_shift = [shift_orig_x, shift_orig_y, shift_orig_z]
        shift_kx = bytes2float(wfk.read(8))
        shift_ky = bytes2float(wfk.read(8))
        shift_kz = bytes2float(wfk.read(8))
        self.kpt_shift = [shift_kx, shift_ky, shift_kz]
        #-----------------------#
        # unpack pseudopotentials
        wfk.read(4)
        self.title:list[str] = []
        self.znuclpsp:list[float] = []
        self.zionpsp:list[float] = []
        self.pspso:list[int] = []
        self.pspdat:list[int] = []
        self.pspcod:list[int] = []
        self.pspxc:list[int] = []
        self.lmn_size:list[int] = []
        self.md5_pseudos:list[str] = []
        for _ in range(self.npsp):
            wfk.read(4)
            self.title.append(wfk.read(132).decode())
            self.znuclpsp.append(bytes2float(wfk.read(8)))
            self.zionpsp.append(bytes2float(wfk.read(8)))
            self.pspso.append(bytes2int(wfk.read(4)))
            self.pspdat.append(bytes2int(wfk.read(4)))
            self.pspcod.append(bytes2int(wfk.read(4)))
            self.pspxc.append(bytes2int(wfk.read(4)))
            self.lmn_size.append(bytes2int(wfk.read(4)))
            self.md5_pseudos.append(wfk.read(32).decode())
            wfk.read(4)
        #---------------#
        # unpack PAW data
        if self.usepaw == 1:
            wfk.read(4)
            self.nrho:list[int] = []
            for _ in range(self.natom):
                for _ in range(self.nspden):
                    nrhoij = bytes2int(wfk.read(4))
                    self.nrho.append(nrhoij)
            self.cplex = bytes2int(wfk.read(4))
            self.nspden_paw = bytes2int(wfk.read(4))
            wfk.read(8)
            self.irho:list[int] = []
            for i in range(self.natom):
                for _ in range(self.nspden_paw):
                    nrhoij = self.nrho[i]
                    for _ in range(nrhoij):
                        rhoij = bytes2int(wfk.read(4))
                        self.irho.append(rhoij)
            wfk.read(4)
            self.rho:list[float] = []
            for i in range(self.natom):
                    for _ in range(self.nspden_paw):
                        nrhoij = self.nrho[i]
                        for _ in range(nrhoij):
                            rhoij = bytes2float(wfk.read(8))
                            self.rho.append(rhoij)
            wfk.read(4)
        print('WFK header read')
        wfk.close()
    #-----------------------------------------------------------------------------------------------------------------#
    # check for time reversal symmetry
    def _CheckTimeRev(
        self
    ):
        # if system is centrosymmetric, do not double reciprocal symmetry operations
        if True in [np.array_equal(-np.identity(3),mat) for mat in self.symrel]:
            self.time_reversal = False
        else:
            self.time_reversal = True
            print((
            'Noncentrosymmetric system identified, assuming time reversal symmetry\n'
            'To change this, set "check_time_rev" keyword to False when calling ReadWFK() or ReadEigenvalues()'
            ))
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read entire body of abinit version 7 wavefunction file
    def ReadWFK(
        self, energy_level=np.nan, width=np.nan, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        '''
        Method that constructs WFK objects from ABINIT v10 WFK file.

        Parameters
        ----------
        energy_level : float
        The energy level to pull plane wave coefficients from
        This is relative to the Fermi energy
        If not defined but width is defined, default is 0 Hartree

        width : float
        Defines range about energy_level to search for plane wave coefficients
        Total range is equal to width, so look width/2 above and below energy_level
        If not defined but energy_level is defined, default is 0.005 Hartree
        '''
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        # read wfk
        wfk = open(self.filename, 'rb')
        skip = True
        if energy_level is np.nan:
            if width is np.nan:
                skip = False
            else:
                energy_level = 0
        else:
            if width is np.nan:
                width = 0.005
        #-----------#
        # skip header
        wfk.read(468)
        wfk.read(4*(2*self.nkpt + self.nkpt*self.nsppol + self.npsp + 10*self.nsym + self.natom))
        wfk.read(8*(4*self.nkpt + self.bandtot + 3*self.nsym + 2*self.ntypat + 3*self.natom))
        wfk.read(self.npsp*(208))
        if self.usepaw == 1:
            wfk.read(4)
            for _ in range(self.natom):
                for _ in range(self.nspden):
                    _ = wfk.read(4)
            wfk.read(4)
            wfk.read(4)
            wfk.read(8)
            for i in range(self.natom):
                for _ in range(self.nspden_paw):
                    nrhoij = self.nrho[i]
                    for _ in range(nrhoij):
                        _ = wfk.read(4)
            wfk.read(4)
            for i in range(self.natom):
                for _ in range(self.nspden_paw):
                    nrhoij = self.nrho[i]
                    for _ in range(nrhoij):
                        _ = wfk.read(8)
            wfk.read(4)
        #-------------------------------#
        # begin reading wavefunction body
        for i in range(self.nsppol):
            for j in range(self.nkpt):
                print(f'Reading kpoint {j+1} of {self.nkpt}', end='\r')
                if j+1 == self.nkpt:
                    print('\n', end='')
                wfk.read(4)
                npw = bytes2int(wfk.read(4))
                self.nspinor = bytes2int(wfk.read(4))
                nband_temp = bytes2int(wfk.read(4))
                pw_indices = np.zeros((npw,3), dtype=int)
                eigenvalues = np.zeros(nband_temp, dtype=float)
                wfk.read(8)
                for pw in range(npw):
                    kx = bytes2int(wfk.read(4))
                    ky = bytes2int(wfk.read(4))
                    kz = bytes2int(wfk.read(4))
                    pw_indices[pw,:] = [kx, ky, kz]
                wfk.read(8)
                for nband in range(nband_temp):
                    eigenval = bytes2float(wfk.read(8))
                    eigenvalues[nband] = eigenval
                # skip reading coefficients if they energy level of interest is not in band
                if skip:
                    min_val = self.fermi + energy_level - width/2
                    max_val = self.fermi + energy_level + width/2
                    eig_found = False
                    for eigval in eigenvalues:
                        if min_val <= eigval <= max_val:
                            eig_found = True
                            break
                    if eig_found:
                        yield self._ReadCoeffs(
                            wfk,
                            nband_temp,
                            npw,
                            eigenvalues,
                            pw_indices,
                            ind=j
                        )
                    else:
                        wfk.read(nband_temp*8)
                        wfk.read(4)
                        wfk.read(nband_temp*(8 + npw*16))
                # if not skip, read full wavefunction
                else:
                    yield self._ReadCoeffs(
                        wfk,
                        nband_temp,
                        npw,
                        eigenvalues,
                        pw_indices,
                        ind=j
                    )
        print('WFK body read')
        wfk.close()
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read in plane wave coefficients
    def _ReadCoeffs(
        self, wfk_file, nbands, npws, eigs, pw_inds, ind
    )->wc.WFK:
        occupancies = np.zeros((1,nbands), dtype=float)
        coeffs = np.zeros((nbands,npws), dtype=complex)
        for nband in range(nbands):
            occ = bytes2float(wfk_file.read(8))
            occupancies[:,nband] = occ
        wfk_file.read(4)
        for nband in range(nbands):
            wfk_file.read(4)
            cg:np.ndarray = np.zeros((1,npws), dtype=complex)
            for pw in range(npws):
                cg1 = bytes2float(wfk_file.read(8))
                cg2 = bytes2float(wfk_file.read(8))
                cg[:,pw] = cg1 + 1j*cg2
            coeffs[nband,:] = cg
            wfk_file.read(4) 
        return wc.WFK(
            eigenvalues=np.array(eigs), 
            wfk_coeffs=np.array(coeffs),
            pw_indices=np.array(pw_inds),
            kpoints=np.array(self.kpts[ind]),
            nkpt=self.nkpt,
            nbands=self.bands[0],
            ngfftx=self.ngfftx,
            ngffty=self.ngffty,
            ngfftz=self.ngfftz,
            symrel=np.array(self.symrel),
            nsym=self.nsym,
            lattice=self.real_lattice,
            natom=self.natom,
            xred=np.array(self.xred),
            typat=self.typat,
            znucltypat=self.znucltypat,
            fermi_energy=self.fermi,
            non_symm_vecs=np.array(self.tnons),
            time_reversal=self.time_reversal
        )
    #-----------------------------------------------------------------------------------------------------------------#
    # method to read only eigenvalues from body of abinit version 10 wavefunction file
    def ReadEigenvalues(
        self, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        '''
        Method that constructs WFK objects from ABINIT v10 WFK file.
        '''
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        # read wfk
        wfk = open(self.filename, 'rb')
        #-----------#
        # skip header
        wfk.read(468)
        wfk.read(4*(2*self.nkpt + self.nkpt*self.nsppol + self.npsp + 10*self.nsym + self.natom))
        wfk.read(8*(4*self.nkpt + self.bandtot + 3*self.nsym + 2*self.ntypat + 3*self.natom))
        wfk.read(self.npsp*(208))
        if self.usepaw == 1:
            wfk.read(4)
            for _ in range(self.natom):
                for _ in range(self.nspden):
                    _ = wfk.read(4)
            wfk.read(4)
            wfk.read(4)
            wfk.read(8)
            for i in range(self.natom):
                for _ in range(self.nspden_paw):
                    nrhoij = self.nrho[i]
                    for _ in range(nrhoij):
                        _ = wfk.read(4)
            wfk.read(4)
            for i in range(self.natom):
                for _ in range(self.nspden_paw):
                    nrhoij = self.nrho[i]
                    for _ in range(nrhoij):
                        _ = wfk.read(8)
            wfk.read(4)
        #-------------------------------#
        # begin reading wavefunction body
        for i in range(self.nsppol):
            for j in range(self.nkpt):
                print(f'Reading kpoint {j+1} of {self.nkpt}', end='\r')
                if j+1 == self.nkpt:
                    print('\n', end='')
                wfk.read(4)
                npw = bytes2int(wfk.read(4))
                self.nspinor = bytes2int(wfk.read(4))
                nband_temp = bytes2int(wfk.read(4))
                eigenvalues = np.zeros((1,nband_temp), dtype=float)
                wfk.read(8)
                wfk.read(12*npw)
                wfk.read(8)
                #------------------------------------------------------------------#
                # only need eigenvalues for Fermi surface, skip over everything else
                for nband in range(nband_temp):
                    eigenval = bytes2float(wfk.read(8))
                    eigenvalues[:,nband] = eigenval
                wfk.read(nband_temp*8)
                wfk.read(4)
                wfk.read(nband_temp*(8 + npw*16))
                yield wc.WFK(
                    eigenvalues=np.array(eigenvalues), 
                    kpoints=np.array(self.kpts[j]),
                    nkpt=self.nkpt,
                    nbands=self.bands[0],
                    ngfftx=self.ngfftx,
                    ngffty=self.ngffty,
                    ngfftz=self.ngfftz,
                    symrel=np.array(self.symrel),
                    nsym=self.nsym,
                    lattice=self.real_lattice,
                    natom=self.natom,
                    xred=np.array(self.xred),
                    typat=self.typat,
                    znucltypat=self.znucltypat,
                    fermi_energy=self.fermi,
                    non_symm_vecs=np.array(self.tnons),
                    time_reversal=self.time_reversal
                )
        print('WFK body read')
        wfk.close()

#---------------------------------------------------------------------------------------------------------------------#
#----------------------------------------------------- END CLASS -----------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#

class AbinitNetCDF():
    def __init__(
        self, filename:str
    ):
        print('WFK found with netCDF4 format')
        self.filename = filename
        self.dataset = self._GetDataset()
        self.kpts:np.ndarray = self.dataset.variables['reduced_coordinates_of_kpoints'][:]
        self.nkpt:int = self.dataset.dimensions['number_of_kpoints'].size
        self.xred:np.ndarray = self.dataset.variables['reduced_atom_positions'][:]
        self.natom:int = self.dataset.dimensions['number_of_atoms'].size
        self.tnons:np.ndarray = self.dataset.variables['reduced_symmetry_translations'][:]
        self.symrel:np.ndarray = self.dataset.variables['reduced_symmetry_matrices'][:]
        self.symrel = np.array([op.T for op in self.symrel])
        self.nsym:int = self.dataset.dimensions['number_of_symmetry_operations'].size
        self.fermi:float = self.dataset['fermi_energy'][:]
        self.real_lattice:np.ndarray = self.dataset['primitive_vectors'][:]
        self.real_lattice *= 0.529177
        self.ngfftx:int = self.dataset.dimensions['number_of_grid_points_vector1'].size
        self.ngffty:int = self.dataset.dimensions['number_of_grid_points_vector2'].size
        self.ngfftz:int = self.dataset.dimensions['number_of_grid_points_vector3'].size
        self.nband:int = int(self.dataset.dimensions['bantot'].size / self.nkpt)
        self.typat:list = self.dataset.variables['atom_species'][:].tolist()
        self.zion:np.ndarray = self.dataset.variables['atomic_numbers'][:]
        self.znucltypat:list = [int(nuc) for nuc in self.zion]
        self.bands = [self.nband]
#---------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------ METHODS ------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#
    # create netCDF dataset
    def _GetDataset(
        self
    ):
        from netCDF4 import Dataset
        return Dataset(self.filename,format='NETCDF4')
    #-----------------------------------------------------------------------------------------------------------------#
    # check for time reversal symmetry
    def _CheckTimeRev(
        self
    ):
        # if system is centrosymmetric, do not double reciprocal symmetry operations
        if True in [np.array_equal(-np.identity(3),mat) for mat in self.symrel]:
            self.time_reversal = False
        else:
            self.time_reversal = True
            print((
            'Noncentrosymmetric system identified, assuming time reversal symmetry\n'
            'To change this, set "check_time_rev" keyword to False when calling ReadWFK() or ReadEigenvalues()'
            ))
    #-----------------------------------------------------------------------------------------------------------------#
    # fetch wavefunction coefficients from netcdf dataset
    def ReadWFK(
        self, energy_level:float=np.nan, width:float=np.nan, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        eigs:np.ndarray = self.dataset.variables['eigenvalues'][:][0]
        pw_inds:np.ndarray = self.dataset.variables['reduced_coordinates_of_plane_waves'][:]
        # restructure wavefunction coefficients
        for kpt in range(self.nkpt):
            print(f'Reading kpoint {kpt+1} of {self.nkpt}', end='\r')
            kpt_coeffs = []
            # format plane wave indices
            kpt_pw_inds = pw_inds[kpt]
            kpt_pw_inds = kpt_pw_inds[~kpt_pw_inds.mask]
            kpt_pw_inds = kpt_pw_inds.reshape((-1,3))
            kpt_pw_inds = np.ma.getdata(kpt_pw_inds)
            coeffs = self.dataset.variables['coefficients_of_wavefunctions'][0,kpt]
            for band in coeffs:
                band = band[0]
                band = band[~band.mask]
                band = band.reshape((-1,2))
                band = np.ma.getdata(band)
                kpt_coeffs.append(band[:,0] + 1j*band[:,1])
            yield wc.WFK(
                eigenvalues=eigs[kpt,:], 
                wfk_coeffs=np.array(kpt_coeffs),
                pw_indices=kpt_pw_inds,
                kpoints=self.kpts[kpt],
                nkpt=self.nkpt,
                nbands=self.nband,
                ngfftx=self.ngfftx,
                ngffty=self.ngffty,
                ngfftz=self.ngfftz,
                symrel=self.symrel,
                nsym=self.nsym,
                lattice=self.real_lattice,
                natom=self.natom,
                xred=self.xred,
                typat=self.typat,
                znucltypat=self.znucltypat,
                fermi_energy=self.fermi,
                non_symm_vecs=self.tnons,
                time_reversal=self.time_reversal
            )
    #-----------------------------------------------------------------------------------------------------------------#
    # fetch eigenvalues from netcdf dataset
    def ReadEigenvalues(
        self, energy_level:float=np.nan, width:float=np.nan, check_time_rev:bool=True
    )->Generator[wc.WFK, None, None]:
        # check for time reversal symmetry
        if check_time_rev:
            self._CheckTimeRev()
        else:
            self.time_reversal = False
        eigs:np.ndarray = self.dataset.variables['eigenvalues'][:][0]
        for i, kpt in enumerate(self.kpts):
            print(f'Reading kpoint {i+1} of {self.nkpt}', end='\r')
            yield wc.WFK(
                eigenvalues=eigs[i,:], 
                kpoints=kpt,
                nkpt=self.nkpt,
                nbands=self.nband,
                ngfftx=self.ngfftx,
                ngffty=self.ngffty,
                ngfftz=self.ngfftz,
                symrel=self.symrel,
                nsym=self.nsym,
                lattice=self.real_lattice,
                natom=self.natom,
                xred=self.xred,
                typat=self.typat,
                znucltypat=self.znucltypat,
                fermi_energy=self.fermi,
                non_symm_vecs=self.tnons,
                time_reversal=self.time_reversal
            )        

#---------------------------------------------------------------------------------------------------------------------#
#----------------------------------------------------- END CLASS -----------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------#

def AbinitWFK(
    filename:str
)->Union[Abinit10WFK, Abinit7WFK, AbinitNetCDF]:
    '''
    A function for identifying ABINIT version and constructing respective AbinitWFK object.
    '''
    if filename.endswith('.nc'):
        return AbinitNetCDF(filename)
    wfk = open(filename, 'rb')
    wfk.read(4)
    version = wfk.read(6).decode()
    wfk.close()
    if version.strip().split('.')[0] == '7':
        return Abinit7WFK(filename)
    elif version.strip().split('.')[0] == '10':
        return Abinit10WFK(filename)
    else:
        print(f'Version {version} may not be supported, attempting to analyze with v10 procedure.')
        return Abinit10WFK(filename)